//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prom_lsc_constr.C,v 1.44 2005/05/19 19:55:17 adams Exp $
//  Author:             Mark F. Adams
//  Filename:           prom_lsc_constr.C
//  April 2002
// 
#include "Prometheus_LinSysCore.h" 

#include "prometheus.hh"
#include "prom_inln.icc" 

#define NOPROM_PRINT if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD, "[?]%s line %d\n", __FUNCT__, __LINE__ )
#define NOPROM_PRINT1(a) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?]%s line %d: %d\n",__FUNCT__,__LINE__,a)
#define NOPROM_PRINTD(a) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?]%s line %d: %e\n",__FUNCT__,__LINE__,a)
#define NOPROM_PRINTS(a) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?]%s line %d: %s\n",__FUNCT__,__LINE__,a)
#define NOPROM_PRINT2(a,b) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?]%s line %d: %d %d\n",__FUNCT__,__LINE__,a,b)
#define NOPROM_PRINT3(a,b,c) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?in an an engineering department]%s line %d: %d %d %d\n",__FUNCT__,__LINE__,a,b,c)
#define NOPROM_PRINT4(a,b,c,d) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?]%s line %d: %d %d %d %d\n",__FUNCT__,__LINE__,a,b,c,d)

#define PRINT if(verbose_>1)PetscPrintf(MPI_COMM_WORLD, "[%d]%s line %d\n",mype_, __FUNCT__, __LINE__ )
#define PRINT1(a) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %d\n",mype_,__FUNCT__,__LINE__,a)
#define PRINTD(a) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %e\n",mype_,__FUNCT__,__LINE__,a)
#define PRINTS(a) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %s\n",mype_,__FUNCT__,__LINE__,a)
#define PRINT2(a,b) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %d %d\n",mype_,__FUNCT__,__LINE__,a,b)
#define PRINT3(a,b,c) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %d %d %d\n",mype_,__FUNCT__,__LINE__,a,b,c)
#define PRINT4(a,b,c,d) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %d %d %d %d\n",mype_,__FUNCT__,__LINE__,a,b,c,d)

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::SymbolicSetUp"
/* Prometheus_LinSysCore::SymbolicSetUp() *************************************
 *
 *  INPUT:
 *
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::SymbolicSetUp() 
{
  PRINT;  assert( prom_->grid0_ != NULL );
  int *logs = PromContext::prom_logs_, matsetup = logs[LOG_FEI_MATRIX_SETUP];
  int promsetup = logs[LOG_FEI_PROM_SETUP],ierr;
  int reg = logs[LOG_PARTITION];
  PromMatrix *KK = prom_->grid0_->stiffness_; assert(KK!=NULL);
  PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK);
  PromMatrix *CC = (use_uzawa_ || K==NULL) ? uzawaC_ : K->C_;

  if( !set_coords_ ) {
    ierr = prom_->grid0_->SetCoords( NULL ); CHKERRQ(ierr);
    set_coords_ = TRUE;
  }

  // make coarse grids and finalize 
  ierr = KK->Assembly();     CHKERRQ(ierr); 
  if( mg_levels_ > 1 ) {
    assert(prom_->grid0_->nextgrid_ == NULL);
    PromGrid *newGrid = NULL, *lastGrid = prom_->grid0_;

    // if coords have not been set, then set default
    const PromAggGrid*const agggrid=dynamic_cast<const PromAggGrid*>(lastGrid);
    if( agggrid != NULL ) {
      ierr = prom_->perf_mon_.EventBegin(promsetup);   CHKERRQ(ierr); 
      prom_->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_PROM]);
      // scale RBMs here!
      ierr = agggrid->ScaleRBM( prom_->right_ ); CHKERRQ(ierr); 
      prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_PROM]);
      ierr = prom_->perf_mon_.EventEnd(promsetup); CHKERRQ(ierr); 
    }

    // make levels
    for( int lev = 1 ; lev < mg_levels_ ; lev++, lastGrid = newGrid ) {
      // make solver object -- " FEI: Prom setup"
      ierr = prom_->perf_mon_.EventBegin(promsetup);   CHKERRQ(ierr); 
      // make new grid, scales fine grid matrix
      ierr = prom_->NewLevel( mg_levels_ - lev - 1 ); CHKERRQ(ierr);
      //
      newGrid = lastGrid->nextgrid_;
      ierr = prom_->perf_mon_.EventEnd(promsetup); CHKERRQ(ierr); 
      if( newGrid == NULL ) break;

      // " FEI: solv.setup"
      ierr = prom_->perf_mon_.EventBegin(logs[LOG_SLES_SETUP]);CHKERRQ(ierr);
      ierr = prom_->perf_mon_.EventBegin(matsetup); CHKERRQ(ierr);
      
      // 1. add regularization (before smoothing)
      if( CC != NULL && use_uzawa_ ) {
	// Get and set regularization (after smoothing, before make solver)
	if( lastGrid->isBase() ) { 
	  MPI_Comm wcomm = KK->map_.comm_.MPIComm(); 
	  int M = CC->getM();
	  if(verbose_>1)PetscPrintf(wcomm,"[%d] Set reg, %d LMs\n",mype_,M);
	  // set reg. " FEI:reg. "
	  ierr = prom_->perf_mon_.EventBegin(reg); CHKERRQ(ierr);
	  ierr = SetReg( KK );                     CHKERRQ(ierr); 
	  ierr = prom_->perf_mon_.EventEnd(reg);   CHKERRQ(ierr);
	  if(verbose_>1)PetscPrintf(wcomm,"[%d] Regularization done\n",mype_);
	}

	// 1.b Add extra adjacs from regularization (and parents, after Prol.)
	ierr = lastGrid->AddExtraGhostsParents( newGrid );
	CHKERRQ(ierr);
      }

      // make P and A <- P'A      
      bool no_sm,gs;
      ierr = lastGrid->options_.HasOptionName("-aggmg_avoid_resmooth",&no_sm);
      CHKERRQ(ierr);
      ierr = lastGrid->options_.HasOptionName("-aggmg_gs_smoothing",&gs);
      CHKERRQ(ierr);
      const PromAggGrid*const lastagggrid=dynamic_cast<const PromAggGrid*>(lastGrid);
      if( KK->map_.isConstNDF() && KK->map_.constNDFSize()==1 && !gs ){
	ierr = prom_->perf_mon_.EventBegin(logs[LOG_RESTRICTION]); 
	CHKERRQ(ierr); assert(newGrid->prolongator_==NULL);
	// 2a. make P0 prolongator
	ierr = newGrid->CreateProlongator( lastGrid, prom_->Comm_ );
	CHKERRQ(ierr);
	// 2b. set P0 prolongator values
	PromMatrix *PP = newGrid->prolongator_; assert(PP!=NULL);
	ierr = newGrid->MakeProlongator( lastGrid, PP, true ); CHKERRQ(ierr);
	ierr = prom_->perf_mon_.EventEnd(logs[LOG_RESTRICTION]); CHKERRQ(ierr);
	// 2c. make P from P0 for smoothed aggregation
	if( lastagggrid != NULL && lastagggrid->amg_smooths_ > 0 ) {
	  // resmooth P, remake prolongator, remake coarse grid operator  
	  ierr = lastGrid->SmoothScalarP( newGrid ); CHKERRQ(ierr); 
	}
	// 3. RAP - create next mat object
	ierr = lastGrid->MakeScalarNextMat( newGrid );  CHKERRQ(ierr);
      }
      else {
	ierr = prom_->perf_mon_.EventBegin(logs[LOG_RESTRICTION]);
	CHKERRQ(ierr); assert(newGrid->prolongator_==NULL);
	// 2a. smooth prolongator (after regularization)
	if( lastagggrid != NULL && lastagggrid->amg_smooths_ > 0 ) {
	  if( gs ){ 
	    ierr = lastagggrid->SmoothP_gs(); CHKERRQ(ierr); 
	  }
	  else{ ierr = lastagggrid->SmoothP();        CHKERRQ(ierr); }
	}
	// 2b. Create P matrix
	ierr = newGrid->CreateProlongator(lastGrid,prom_->Comm_);CHKERRQ(ierr);
	// 3a. create matrices
	assert(newGrid->stiffness_==NULL);
	ierr = newGrid->MakeMatrix( use_complex_, (K != NULL), 
				    (CC!=NULL && use_uzawa_)? 15:0);
	CHKERRQ(ierr);
	// 3b. RAP - create next mat object
	ierr = lastGrid->MakeNextMat( newGrid ); CHKERRQ(ierr);
	// 2c. set prolongator values (deletes 'DAP' - used in RAP - when resmoothing)
	PromMatrix *PP = newGrid->prolongator_; assert(PP!=NULL);
	ierr = newGrid->MakeProlongator(lastGrid, PP, !no_sm );CHKERRQ(ierr);
	ierr = prom_->perf_mon_.EventEnd(logs[LOG_RESTRICTION]); CHKERRQ(ierr);
      }
      ierr = prom_->perf_mon_.EventEnd(matsetup); CHKERRQ(ierr);
      ierr = prom_->perf_mon_.EventEnd(logs[LOG_SLES_SETUP]); CHKERRQ(ierr);
      
      if( CC != NULL && !use_uzawa_ ) {
	assert(lastGrid->nextgrid_ == newGrid);
	// 4. all-at-once, after next grid - does R A P for constraints
	ierr = CreateKKTObs( lastGrid ); CHKERRQ(ierr);  
      }

      if( CC != NULL && !use_uzawa_ && verbose_ > 2 ) { // debug
	//
	FILE *file; char path[128]; assert(K!=NULL);
	if( npe_ == 1 ) sprintf( path, "%s/K_%d", root_, lev+1 );
	else sprintf( path, "%s/K_%d_%d", root_, lev+1, mype_ );
	file = fopen( path, "w");
	ierr = PromContext::WriteMat_private(K,file,0,0,false);CHKERRQ(ierr);
	ierr = PromContext::WriteMat_private(K->C_,file,K->getM(),0,false);
	CHKERRQ(ierr);
	ierr = PromContext::WriteMat_private(K->C_,file,0,K->getM(),true);
	CHKERRQ(ierr);
	fclose(file);
      }
    }
    // top grid 'lastGrid'
    if( CC != NULL && !use_uzawa_ ) {
      assert(lastGrid->nextgrid_ == NULL);
      // all-at-once, after next grid
      ierr = CreateKKTObs( lastGrid ); CHKERRQ(ierr);
    }
  }
  else {  // first time, one level solver
    if( CC != NULL ) {
      // " FEI: Prom setup"
      ierr = prom_->perf_mon_.EventBegin(matsetup);    CHKERRQ(ierr); 
      if( use_uzawa_ ) { 
	// set reg. " FEI: reg."
	ierr = prom_->perf_mon_.EventBegin(reg); CHKERRQ(ierr);
	ierr = SetReg( KK ); CHKERRQ(ierr); 
	ierr = prom_->perf_mon_.EventEnd(reg); CHKERRQ(ierr);
      }
      else{
	ierr = CreateKKTObs( prom_->grid0_ ); CHKERRQ(ierr);
      }
      ierr = prom_->perf_mon_.EventEnd(matsetup); CHKERRQ(ierr);
    }
  }
  
  // finalize
  assert( !prom_->finalized_ );
  
  // " FEI: Prom setup"
  ierr = prom_->perf_mon_.EventBegin(promsetup); CHKERRQ(ierr);
  ierr = prom_->FinalizeSymbolic(); CHKERRQ(ierr);
  ierr = prom_->perf_mon_.EventEnd(promsetup); CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::MakeSolver"
/* Prometheus_LinSysCore::MakeSolver() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::MakeSolver( const PromMap *const p_map, bool ispetscpc ) 
{
  PRINT;
  int ierr,maxit; PromGrid *grid0 = prom_->grid0_; 
  PromMatrix *KK = grid0->stiffness_; bool flag;
  const PromMap &map = KK->map_;

  PetscFunctionBegin;
  // make coarse grid operators and sets up solver (matrix scaled)
  const int numlevs = prom_->getNumLevels();
  // set parameters and alloc
  assert(prom_->solver_==NULL);
  // all at once solve if p_map != NULL
  prom_->solver_ =  new PromSolver(prom_->options_,prom_->perf_mon_,map,p_map);
  CHKNEW(prom_->solver_,1);

  ierr = prom_->options_.GetOptionInt("-ksp_max_it",&maxit,&flag); CHKERRQ(ierr);
  if( !flag ) maxit = 200;
  else if( maxit <= 0 ) maxit = 20;

  // make solver
  if( numlevs > 1 ) {
    ierr = prom_->solver_->CreatePC( PROMPCMG, xVec_ ); CHKERRQ(ierr);
    if( !ispetscpc ) ierr = prom_->solver_->SetKSP( maxit ); 
    else ierr = prom_->solver_->SetKSP( 1, NULL, "preonly" );
    CHKERRQ( ierr );
    ierr = prom_->solver_->SetOperator( KK ); CHKERRQ(ierr); // restes
    PromPCMG *const mgpc = dynamic_cast<PromPCMG*>(prom_->solver_->pc_);
    assert(mgpc!=NULL); 
    // create MG solver objects -- symmetrizes matrix
    ierr = mgpc->Create( grid0, p_map ); CHKERRQ(ierr);
    assert(mgpc->getNumIts() == 1);
  }
  else { // one grid
    // assemble fine grid -- done in 'newLevel' (scales matrix)
    ierr = prom_->SetMatrixWithNewData( KK ); CHKERRQ(ierr);
    bool kspset = false;
    ierr = prom_->solver_->MakePC( grid0, grid0->nBlocks_,
				   grid0->blockISs_, kspset, xVec_ );
    CHKERRQ(ierr);
    if( !kspset ) {
      if( !ispetscpc ) ierr = prom_->solver_->SetKSP( maxit ); 
      else ierr = prom_->solver_->SetKSP( 1, NULL, "preonly" ); 
      CHKERRQ(ierr);
    }
    else assert(!ispetscpc);
    // setup KKT PC
    if( p_map != NULL ) {
      assert(xVec_ != NULL);  PromMatrix *KK = grid0->stiffness_; 
      PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK); assert(K!=NULL);
      PromCRVector *X = dynamic_cast<PromCRVector*>(xVec_); assert(X!=NULL);
      PromPC_base *pc = prom_->solver_->pc_; assert(pc!=NULL);
      ierr = pc->CreateKKTSmoother( KK, K->C_, X, grid0->nCRAggs_,
				    grid0->CRAggs_, K->CDinvCt_ );
      CHKERRQ(ierr);
      grid0->CRAggs_ = NULL;  grid0->nCRAggs_ = 0; // takes these
      ierr = pc->kkt_->SetUp(); CHKERRQ(ierr); // same state as others
    }
  }
  
  // convert to symm
  for( PromGrid *grid = grid0 ; grid != NULL ; grid = grid->nextgrid_ ) {
    if( !grid->isTop() || grid->isBase() ) {
      ierr = prom_->ConvertToSymm( grid->stiffness_, prom_->options_ );
      CHKERRQ(ierr);
    }
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::ReFactorSolver"
/* Prometheus_LinSysCore::ReFactorSolver() ************************************
 *
 *  - reloaded just constraint equations, called on 2+ solves
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 *    - 
 */
int Prometheus_LinSysCore::ReFactorSolver() 
{
  PRINT;
  int ierr,*logs=PromContext::prom_logs_,res=logs[LOG_RESTRICTION];
  const int reg=logs[LOG_PARTITION];
  const int matsetup=logs[LOG_FEI_MATRIX_SETUP];
  PromMatrix *KK = prom_->grid0_->stiffness_;
  PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK); 
  bool haveCR = (lmap_!=NULL); MPI_Comm comm = MPIComm_;
  PromSolver *solver = prom_->solver_;
  PromOptions &opts = prom_->options_;

  // assemble fine grid -- used on solves 1+ solves (scales matrix)
  ierr = prom_->SetMatrixWithNewData( KK ); CHKERRQ(ierr);

  // just re"factor"
  if( !haveCR ) { // no CRs
    int nn; bool flag; 
    ierr = opts.GetOptionInt("-prometheus_reuse_matrix_factors",&nn,&flag);
    CHKERRQ(ierr);
    if( flag ) {
      if(matrix_setup_reuse_factor_ != nn) num_matrices_ = 0; // new/first one
      matrix_setup_reuse_factor_ = nn;
    }
  }
  // 
  num_matrices_++;
  if( num_matrices_ % matrix_setup_reuse_factor_ != 0 ) {
    if(verbose_>0)PetscPrintf(comm,"\t[0]%s reuse matrix data\n",__FUNCT__);
  }
  else {
    if(verbose_>1)PetscPrintf(comm,"[0]%s redo matrix 'factors'\n",__FUNCT__);
    ierr = prom_->perf_mon_.EventBegin(matsetup); CHKERRQ(ierr);
    // remake coarse grids
    const int numlevs = prom_->getNumLevels(); 
    if( numlevs > 1 ) {
      PromPCMG *mgpc = dynamic_cast<PromPCMG*>(solver->pc_);
      assert(mgpc!=NULL);
      // make matrices -- this is done on grid setup so must be done here!
      PromMG *mg = mgpc->mgstack_, *cmg; 
      for( PromGrid *grid = prom_->grid0_ ; mg != NULL ; 
	   mg = cmg, grid = grid->nextgrid_ ) {
	cmg = mg->next_; 
	PromGrid *nextg = grid->nextgrid_;
	// 
	if( verbose_ > 1 ) {
	  prom_->PrintMatInfo( grid->stiffness_, grid->getLevel(), 
			       grid->factor_, mype_, npe_ );
	}	
	// 1. regularization (before smoothing) -- not used
	if( haveCR && use_uzawa_ ) {
	  assert(0); // reuse of constraint data not used!
	  if( grid->isBase() ) { 
	    if(verbose_>1)PetscPrintf(comm,"[%d] Set reg, %d LMs\n",
				      mype_,uzawaC_->getM());
	    // -- " FEI: reg."
	    ierr = prom_->perf_mon_.EventBegin(reg);  CHKERRQ(ierr);
	    ierr = SetReg( KK ); CHKERRQ(ierr);
	    ierr = prom_->perf_mon_.EventEnd(reg); CHKERRQ(ierr);
	    if(verbose_>1)PetscPrintf(comm,"[%d] Regularization done\n",mype_);
	  }
	  
	  // 1.b add next ghost pars (uzawa)
	  if( nextg != NULL ) {
	    // Add extra adjacs from regularization (and parents, after Prol.)
	    ierr = grid->AddExtraGhostsParents( nextg ); CHKERRQ(ierr);
	    // delete next as 
	    PromMatrix *A = nextg->stiffness_; assert(A!=NULL);
	    float nloc=(float)nextg->nlocal(), nnzloc=(float)A->getLocalNNZ();
	    float nloceq=A->getm()+.01;
	    int nz=(int)((nnzloc/nloceq)*(nloc/nloceq) + 3.0);
	    if( nz > nloc ) nz = (int)nloc + 1;
	    else if( nloc == 0 ) nz = 0;
	    delete nextg->stiffness_; // not epetra complient!
	    nextg->stiffness_ = new PromCRMatrix(*nextg, nz, NULL, nz/2, NULL);
	  }
	}
	
	// 1.a create coarsening and setup KKT PC - does R A P for constraints
	PromPC_base *aao_pc = NULL;
	if( haveCR && !use_uzawa_ ) {
	  aao_pc = mg->presmooth_->pc_;
	  assert( aao_pc->kkt_ == NULL ); assert(grid->CRmap_ != NULL);
	  ierr = CreateKKTObs( grid ); CHKERRQ(ierr); 
	  assert(grid->CRmap_ != NULL);
	  // setup new P vecs
	  ierr = mg->SetNewPVec( *grid->CRmap_ ); CHKERRQ(ierr);
	  assert(aao_pc->kkt_ == NULL);
	}

	if( cmg != NULL ) {
	  bool gs,no_sm = (grid->getMGType() != PROM_AGGMG);
	  if( !no_sm ) opts.HasOptionName( "-aggmg_avoid_resmooth", &no_sm );
	  if( !no_sm ) {
	    // resmooth P, remake prolongator, remake coarse grid operator  
	    const PromAggGrid*const lastagggrid=dynamic_cast<const PromAggGrid*>(grid);
	    if( lastagggrid != NULL && lastagggrid->amg_smooths_ > 0 ) {
	      ierr = prom_->perf_mon_.EventBegin(res); CHKERRQ(ierr);
	      // 2. (re)smooth (after regularization)
	      ierr=opts.HasOptionName("-aggmg_gs_smoothing",&gs); CHKERRQ(ierr);
	      if( KK->map_.isConstNDF() && KK->map_.constNDFSize()==1 && !gs ){
		// have to delete old matrix - it has fill from smoothing!
		assert(nextg->prolongator_!=NULL && cmg->restrict_==nextg->prolongator_);
		delete nextg->prolongator_; nextg->prolongator_ = NULL;
		// 2a. form P0 matrix
		ierr = nextg->CreateProlongator( grid, prom_->Comm_ ); 
		CHKERRQ(ierr);
		PromMatrix *PP = nextg->prolongator_; assert(PP!=NULL);
		cmg->interpolate_ = cmg->restrict_ = PP; // hack!
		// 2b. set prolongator (P_0) values
		ierr = nextg->MakeProlongator( grid, PP ); CHKERRQ(ierr);
		// 2c. smooth P0 and make P from P0
		ierr = grid->SmoothScalarP( nextg ); CHKERRQ(ierr); 
	      }
	      else {
		PromMatrix *PP = nextg->prolongator_; assert(PP!=NULL);
		// 2a. create (smoothed) P data
		if( gs ){ ierr = lastagggrid->SmoothP_gs(); CHKERRQ(ierr);}
		else{ ierr = lastagggrid->SmoothP();        CHKERRQ(ierr);}
		// 2b. make P
		ierr = nextg->MakeProlongator( grid, PP );  CHKERRQ(ierr); 
	      }
	      ierr = prom_->perf_mon_.EventEnd( res ); CHKERRQ(ierr);
	    }
	  }

	  // 3. RAP - create next mat object
	  if( KK->map_.isConstNDF() && KK->map_.constNDFSize()==1 ){
	    ierr = grid->MakeScalarNextMat( nextg );  CHKERRQ(ierr);
	  }
	  else {
	    ierr = grid->MakeNextMat( nextg ); CHKERRQ(ierr);
	  }
	  //
	  ierr = mg->SetOperator( grid->stiffness_ );   CHKERRQ(ierr);

	  // setup coarse grids
	  PromMatrix *nA = nextg->stiffness_;  assert(cmg->presmooth_ != NULL);
	  if( haveCR && use_uzawa_ ) {
	    // set operators (sets dirty bit)
	    ierr = cmg->SetNewOperator( nA );   CHKERRQ(ierr);
	  }
	  else {
	    ierr = cmg->SetOperator( nA );   CHKERRQ(ierr);
	    if( aao_pc != NULL ) {
	      assert( aao_pc->kkt_ == NULL ); 
	      // create new smoother 
	      PromMatrix *KK = grid->stiffness_; 
	      PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK); assert(K!=NULL);
	      PromCRVector *xx = dynamic_cast<PromCRVector*>(mg->r_); 
	      assert(xx!=NULL);
	      ierr = aao_pc->CreateKKTSmoother(K, K->C_, xx, grid->nCRAggs_,
					       grid->CRAggs_, K->CDinvCt_ );
	      CHKERRQ(ierr);
	      grid->CRAggs_ = NULL; grid->nCRAggs_ = 0; // takes these 
	      assert(aao_pc->setup_called_==2);         // skip around
	      ierr = aao_pc->kkt_->SetUp(); CHKERRQ(ierr); // keep in same state as rest - noop!
	    }
	  }
	}
	else { // top grid 
	  if( haveCR && !use_uzawa_ ) {
	    // create new smoother 
	    PromPC_base *pc = mg->presmooth_->pc_;
	    if( pc->kkt_ == NULL ) {
	      PromCRVector *xx = dynamic_cast<PromCRVector*>(mg->x_); 
	      assert(xx!=NULL); const PromMatrix *A = grid->stiffness_;
	      const PromCRMatrix *K = dynamic_cast<const PromCRMatrix*>(A);
	      const PromMatrix *C = K->C_; 
	      ierr = pc->CreateKKTSolver( A, C, xx ); CHKERRQ(ierr);
	      assert(grid->CRAggs_ == NULL); assert(grid->nCRAggs_ == 0); 
	    }
	  }
	  ierr = mg->SetOperator( grid->stiffness_ );   CHKERRQ(ierr);
	}
      }
    }
    else{ // for one level solve
      if ( haveCR ) {
	if( use_uzawa_ ) {
	  // -- " FEI: reg."
	  ierr = prom_->perf_mon_.EventBegin(reg); CHKERRQ(ierr); assert(K!=NULL);
	  ierr = SetReg( KK ); CHKERRQ(ierr);  // set regularization, 
	  ierr = prom_->perf_mon_.EventEnd(reg); CHKERRQ(ierr);
	}
	else {
	  // does little
	  PromGrid *grid0 = prom_->grid0_;
	  ierr = CreateKKTObs( grid0 ); CHKERRQ(ierr);
	  ierr = solver->SetNewPVec( *grid0->CRmap_ ); CHKERRQ(ierr);
	  // create new smoother
	  PromCRVector *X = dynamic_cast<PromCRVector*>(xVec_); assert(X!=NULL);
	  PromPC_base *pc = solver->pc_; assert(pc!=NULL); 
	  ierr = pc->CreateKKTSmoother( K, K->C_, X, grid0->nCRAggs_,
					grid0->CRAggs_, K->CDinvCt_ );
	  CHKERRQ(ierr);
	  grid0->CRAggs_ = NULL;  grid0->nCRAggs_ = 0; // takes these
	}
      }
    }
    // set operators
    if( haveCR && use_uzawa_ ) {
      // regularization can change
      ierr = solver->SetNewOperator( KK ); CHKERRQ(ierr); 
    }
    else {
      ierr = solver->SetOperator( KK ); CHKERRQ(ierr); // could set kkt
      if( haveCR ) {
	assert(!use_uzawa_); // aao mg
	ierr = solver->SetNewPVec( K->C_->map_ ); CHKERRQ(ierr); 
      }
    }
    // solver setup
    ierr = solver->SetUp(); CHKERRQ(ierr);

    // 3.b check for symm
    for( PromGrid *grid = prom_->grid0_ ; grid != NULL ; 
	 grid = grid->nextgrid_ ) {
      if( !grid->isTop() && !grid->isBase() ) {
	ierr = prom_->ConvertToSymm( grid->stiffness_, opts ); CHKERRQ(ierr);
      }
    }
    ierr = prom_->perf_mon_.EventEnd(matsetup); CHKERRQ(ierr);
  }
 
  return 0;
}


#undef __FUNCT__    
#define __FUNCT__ "PromUzawaSolver::SetReg"
/* PromUzawaSolver::SetReg() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int PromUzawaSolver::SetReg( PromMatrix *AA ) 
{
  int ierr;
  PromMatrix *CC = uzawaC_;  assert(CC!=NULL);

  // A <- A + alpha C' * P * C, getrow on C, add on A, get diag on A
  ierr = AA->GetDiagonal( diag_ );      CHKERRQ(ierr);
  // add regularization
  ierr = setRegularization_private(AA, CC, diag_, penalty_factor_, Pen_stiff_);
  CHKERRQ(ierr);
  ierr = AA->Assembly(); CHKERRQ(ierr);
  AA->SetDirty(); // not all procs set anything, so do manually

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "PromUzawaSolver::setDummyReg"
/* PromUzawaSolver::setDummyReg() ***********************
 *
 *  - 
 *
 *  INPUT: 
 *   - AA: stiffness matrix
 *   - CR_id: 
 *
 *  SIDE EFFECTS:
 */
int PromUzawaSolver::setDummyReg( PromMatrix *const AA, 
				  const PromTable<PromCR*> *const CR_id )
  const
{
  // add fake reg. to fool PETSc
  int ii,jj,ierr,bsz=PROM_MAXNDF*PROM_MAXNDF,buffsz=0,proc,tt;
  const int *p_gnode = AA->map_.proc_gnode(),np=AA->map_.comm_.np();
  double zeros[16384]; // 128^2
  const int mype = AA->map_.comm_.mype(); const int tag = AA->map_.comm_.getNewTag();
  const int my0 = p_gnode[mype], next0 = p_gnode[mype+1];
  const int reqsz = sizeof(MPI_Request)/sizeof(int); 
  MPI_Comm wcomm = AA->map_.comm_.MPIComm();
  PromList<MPI_Request*> smessages; 
  MPI_Status    status; 
  PromTable<PromList<int>*> proc_list(57); PromList<int> *list; 
  int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 

  TablePos tpos = CR_id->getHeadPosition();
  while( tpos ) {
    const PromCR *const mcr = CR_id->getNext( tpos ); 
    int *pi = mcr->gnodes_, len = mcr->len(), sz = len*bsz; 
    if(sz>buffsz){
      assert(sz<16384); 
      for(ii=buffsz;ii<sz;ii++) zeros[ii]=0.0;
      buffsz = sz; 
    }
    for(ii=0;ii<len;ii++){
      int gidi = pi[ii], proc;
      if( gidi < my0 || gidi >= next0 ) {
	PromGrid::FindProc(p_gnode, np, gidi, proc); assert(proc!=mype);
	if( (list=proc_list.find(proc+1)) == NULL ) {
	  list = new PromList<int>(100); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	}
	list->addTail( gidi+1 ); list->addTail( len + 1 ); 
	for( jj = 0 ; jj < len ; jj++ ) list->addTail( pi[jj] + 1 );
      }
      else {
	ierr = AA->SetValuesBlocked( 1, &gidi, len, pi, zeros, ADD_VALUES );
	CHKERRQ(ierr);
      }
    }
  }

  // send(tag) *< gid, nj, nj*<gidj> >
  for( ii = 0 ; ii < np ; ii++ ) sbuf[ii] = 0;  
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    int tt = list->getCount(); 
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq);
    CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) *pb++ = list->removeHead() - 1;  
    if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
    // send(tag) *< gid, nj, nj*<gidj> >
    MPI_Isend( buff, tt, MPI_INT, proc, tag, wcomm, sreq );
    delete list;
    sbuf[proc] = tt; assert(tt>0);
  }

  // recive requests -- global!
  MPI_Alltoall( sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, wcomm );
  for( proc = 0 ; proc < np ; proc++ ) {
    if( rbuf[proc] != 0 ) {
      // recv(tag)
      int tt = rbuf[proc], *buff, *pb, __buff[PROM_BSZ]; 
      if( tt > PROM_BSZ ) {
	ierr = PetscMalloc(tt*sizeof(int),&buff );  CHKERRQ(ierr); pb = buff;
      }
      else pb = buff = __buff;
      // recv(tag) *< gid, nj, nj*<gidj> >
      MPI_Recv( buff, tt, MPI_INT, proc, tag, wcomm, &status );
      MPI_Get_count( &status, MPI_INT, &ierr ); assert(tt==ierr);
      proc = status.MPI_SOURCE;
      while( pb-buff < tt ) {
	int gidi = *pb++; assert(gidi >= my0 && gidi < next0);
	int len = *pb++, sz = len*bsz;
	if(sz>buffsz){
	  assert(sz<16384); 
	  for(ii=buffsz;ii<sz;ii++) zeros[ii]=0.0;
	  buffsz = sz; 
	}
	ierr = AA->SetValuesBlocked( 1, &gidi, len, pb, zeros, ADD_VALUES );
	CHKERRQ(ierr);
	pb += len;
      }
      assert(pb-buff == tt);   if(buff != __buff) PetscFree(buff); 
    }
  }
  // cleanup sends 
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe( MPI_ANY_SOURCE, tag, wcomm, &tt, &status );   assert(!tt);

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "PromUzawaSolver::setRegularization_private"
/* PromUzawaSolver::setRegularization_private() ***********************
 *
 *  - 
 *
 *  INPUT: 
 *   - AA: stiffness matrix, had C' * dd * C added to it (in/out)
 *   - C: 
 *   - diag: diagonal of AA
 *   - alpha: regualrization scaling
 *   - dd: regualrization scaling array = alpha * average-diag[] (out)
 *
 *  SIDE EFFECTS:
 */
int PromUzawaSolver::setRegularization_private( PromMatrix *const AA, 
						const PromMatrix *const CC, 
						PromVector *const diag,
						const double alpha,
						PromVector *const dd/*=NULL*/ )
{
  int  ierr,ii,jj,eq,primeq,loceq,lmeq,*colids,ndiag,maxcols=22;
  int *ids,*mask;
  const int my0Lm   = CC->map_.my0Eq(), nlocLms = CC->getm();
  const int my0prim = AA->map_.my0Eq(), nlocPrs = AA->getm();
  double scale, *vv,*diagv=NULL,*buffer,*vp; double norm=1.0;
  
  ierr = diag->Norm2( &norm ); CHKERRQ(ierr);
  ierr = diag->GetArray( &diagv );    CHKERRQ(ierr);
  ierr = PetscMalloc(sizeof(double)*(maxcols*(2+maxcols)),&buffer); 
  CHKERRQ(ierr);
  ids = (int*)(buffer+maxcols*maxcols);  mask = ids + maxcols;
  if(CC->getm() != nlocLms)SETERRQ2(1,"C.m(%d) != nlocLms(%d)",
				    CC->getm(),nlocLms);
  for( eq = 0, lmeq = my0Lm ; eq < nlocLms ; eq++, lmeq++ ){
    ierr = CC->GetRow( lmeq, &jj, &colids, &vv ); CHKERRQ(ierr);
    const int ncols = jj;
    if( ncols > maxcols ){
      PetscFree(buffer);   maxcols = ncols + 3;
      ierr = PetscMalloc(sizeof(double)*(maxcols*(2+maxcols)),&buffer); 
      CHKERRQ(ierr);
      ids = (int*)(buffer + maxcols*maxcols); mask = ids + maxcols; 
    }
    for( ii = jj = 0, scale = 0.0, ndiag = 0 ; ii < ncols ; ii++ ) {
      primeq = colids[ii];
      if( vv[ii] != 0.0 ) { // this enforces BCs as C is zeroed out
	ids[jj] = primeq; 
	loceq = primeq - my0prim;
	mask[ii] = 1;
	if( loceq >= 0 && loceq < nlocPrs ) { 
	  scale += diagv[loceq]; ndiag++; 
	}
	jj++;
      }
      else mask[ii] = 0;
    }
    const int ncolids = jj;
    if( ndiag == 0 || scale == 0.0 ) {
      scale = alpha*norm; // wild-ass-guess
      if(ncolids!=0)
	PetscPrintf(PETSC_COMM_SELF,
		    "[?]Warning: constraint eq. %d has no local diags (ndiag=%d) ncolids=%d scale=%e\n",
		    lmeq,ndiag,ncolids,scale);
    }
    else scale = alpha*(scale/(double)ndiag); // average of diagonal K
    if( ncolids != 0 ) {
      // regularize A += scale * vv * vv'
      for( vp = buffer, ii = 0 ; ii < ncols ; ii++ ) {
	if( mask[ii] != 0 ) {
	  const double ti = scale * vv[ii];
	  for(jj=0;jj<ncols;jj++) {
	    if( mask[jj] != 0 ) *vp++ = ti * vv[jj];
	  }
	}
      }
      // add in
      ierr = AA->SetValues(ncolids, ids, ncolids, ids, buffer, ADD_VALUES);
      CHKERRQ(ierr);
    }
    if( dd != NULL ) {
      // make panalty "matrix" for Uzawa
      ierr = dd->SetValues(1, &lmeq, &scale, INSERT_VALUES);  CHKERRQ(ierr);
    }
    jj = ncols;
    ierr = CC->RestoreRow( lmeq, &jj, &colids, &vv ); CHKERRQ(ierr);
  }
  ierr = diag->RestoreArray( &diagv );    CHKERRQ(ierr);
  PetscFree(buffer);

  return 0;
}

/************************************************************************
 * Create KKT PC
 ************************************************************************/
#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::CreateKKTObs"
/* Prometheus_LinSysCore::CreateKKTObs() ***************************
 *
 *  - creates new CRs for coarse grid (if appl.) 
 *  - aggregate constraints and create coarse grid constraints 'C_'
 *  - create 'D_' projector of 'prolongator_'
 *
 *  INPUT: 
 *   - grid: CR aggregates cached in 'CRAggs_' and 'nCRAggs_'
 *
 *  SIDE EFFECTS:
 */  
int Prometheus_LinSysCore::CreateKKTObs( PromGrid*const grid )
{
  int ierr,ii,jj; const PromComm &wcomm = prom_->Comm_;
  PromGrid *const cgrid = grid->nextgrid_; // can be null
  const PromMatrix *const KK = grid->stiffness_;
  const PromCRMatrix *K = dynamic_cast<const PromCRMatrix*>(KK); 
  PRINT;
  ierr = prom_->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_AUX2]); 
  CHKERRQ(ierr);
  // make next 
  PetscFunctionBegin;
  if( !grid->isTop() || grid->isBase() ) {
    assert(K!=NULL); // make coarse CRs
    const PromMatrix *const CC=K->C_, *const Ct = K->Ct_; assert(CC!=NULL);
    // make coarse CRs
    PromMatrix *Ti = NULL; PromVector *tdiag = NULL;
    if( !grid->isTop() ) { // else one level solve, make CC'
      // make CP := C P = ( P' C' )'
      if( cgrid->scalarMap_ == NULL ){
	cgrid->scalarMap_ = new PromMap( cgrid->nlocalEq(), 1, wcomm );
      }
      if(verbose_>2)PetscPrintf(wcomm.MPIComm(),"\t[0]%s C * P\n",__FUNCT__);
      PromMatrix *CP, *PtCt; assert(cgrid->prolongator_!=NULL);
      ierr = TransMult( cgrid->prolongator_, Ct, cgrid->scalarMap_, &PtCt,
			&CC->map_, &CP );
      CHKERRQ(ierr);
      // form aggregates for CRs, scale Ti
      ierr = prom_->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FEAP]);
      CHKERRQ(ierr);
      // T = CP * CP'
      if(verbose_>2)PetscPrintf(wcomm.MPIComm(),"\t[0]%s CP*CP'\n",__FUNCT__);
      ierr = TransMult( PtCt, PtCt, &CC->map_, &Ti ); CHKERRQ(ierr);
      ierr = Ti->Shift( 1.e-100 );                            CHKERRQ(ierr);
      tdiag = new PromVector( Ti->map_ );
      ierr = Ti->GetDiagonal( tdiag );                       CHKERRQ(ierr);
      ierr = tdiag->Reciprocal();                            CHKERRQ(ierr);
      ierr = tdiag->Sqrt();                                  CHKERRQ(ierr);
      ierr = Ti->Scale( tdiag, tdiag );                      CHKERRQ(ierr);
      ierr = prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FEAP]);
      CHKERRQ(ierr);
      
      if( verbose_ > 2 ){
	FILE *file;char path[128]; 
	if( npe_ == 1 ) sprintf(path,"%s/T_%d",root_,grid->getLevel());
	else sprintf(path,"%s/T_%d_%d",root_,grid->getLevel(),mype_);
	file = fopen(path,"w"); 
	PromContext::WriteMat_private(Ti,file,0,0,false); 	
	fclose(file);
      }

      if(verbose_>2)PetscPrintf(wcomm.MPIComm(),"\t[0]%s Agg CR\n",__FUNCT__);
      // form CR prolongator: aggregate with T, scale Ti
      PromMap *clmap; PromMatrix *prol; 
      PromList<PromAList<int>*> aggs;
      // AggregateCRs(): aggregate constraints
      ierr = AggregateCRs( aggs, Ti, 0.2, true ); CHKERRQ(ierr);
      //ierr = AggregateCRs2( aggs, Ti, 0.9, 0.2, true ); CHKERRQ(ierr);
      // move aggs to better procs
      if(verbose_>2)PetscPrintf(wcomm.MPIComm(),"\t[0]%s Move\n",__FUNCT__);
      ierr = MoveAggs( Ti->map_, aggs, cgrid->factor_, &PtCt->map_, CP);
      CHKERRQ(ierr);
      // make objects
      if(verbose_>2)PetscPrintf(wcomm.MPIComm(),"\t[0]%s p Map\n",__FUNCT__);
      ierr = MakeCRProlMap( aggs, grid, &clmap, &prol ); CHKERRQ(ierr);
      cgrid->CRmap_ = clmap;
      PromCRMatrix *P = dynamic_cast<PromCRMatrix*>(cgrid->prolongator_); 
      assert(P!=NULL); P->D_ = prol;
      assert(aggs.isEmpty()); // delete 'aggs'

      // Make C_{i+1} := p' * CP (= (CP)' * p)'
      PromMatrix *nextC, *nextCt;
      if( cgrid->scalarMap_ == NULL ){
	cgrid->scalarMap_ = new PromMap( cgrid->nlocalEq(), 1, wcomm );
      }
      if(verbose_>2)PetscPrintf(wcomm.MPIComm(),"\t[0]%s p'*CP\n",__FUNCT__);
      ierr = TransMult( prol, CP, clmap, &nextC, cgrid->scalarMap_, &nextCt );
      CHKERRQ(ierr);
      PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(cgrid->stiffness_); assert(K!=NULL);
      assert(K->C_==NULL && K->Ct_==NULL);
      K->C_ = nextC; K->Ct_ = nextCt;
      delete PtCt; delete CP;
    }
    // save aggs as ISs for solve
    if(verbose_>2)PetscPrintf(wcomm.MPIComm(),"\t[0]%s CR smooth\n",__FUNCT__);
    assert(grid->CRAggs_==NULL); bool flag; char str[32];
    prom_->options_.GetOptionString( "-cr_pc_type", str, 32, &flag );
    if( !flag || strcmp(str,"asm")==0 || (strcmp(str,"ilu")==0) ) {
      PromList<PromAList<int>*> aggs; // make 'grid->CRAggs_'
      if( (Ti==NULL) || (strcmp(str,"ilu")==0) || CC->getm() < 1000 ){ 
	PromAList<int> *crlist = new PromAList<int>(15);
	aggs.addTail( crlist ); int nn = CC->map_.nlocalEq();
	for(ii=0,jj=CC->map_.my0Eq(); ii<nn; ii++,jj++) crlist->addTail(jj+1);
      }
      else { // aggregate -- collective!!!
	ierr = AggregateCRs( aggs, Ti, 2.e-1, false, true ); CHKERRQ(ierr);
      }
      if( aggs.getCount() > 0 ) {
	int naggs = aggs.getCount(),buff__[PROM_BSZ],bsz=PROM_BSZ,*buf,kk,nn;
	ierr = PetscMalloc(sizeof(PromIS)*naggs, &grid->CRAggs_ ); 
	CHKERRQ(ierr);
	for( ii = nn = 0, buf = buff__ ; !aggs.isEmpty() ; ii++ ) {
	  PromList<int> *crlist = aggs.removeHead(); kk = 0;
	  if( crlist->getCount() > bsz ) {
	    if( buf != buff__ ) PetscFree(buf);
	    bsz = crlist->getCount();
	    ierr = PetscMalloc(sizeof(int)*bsz, &buf ); CHKERRQ(ierr);
	  }
	  while( !crlist->isEmpty() ){ 
	    int id = crlist->removeHead() - 1; 
	    buf[kk++] = id; // local IDs
	  } 
	  nn += kk;
	  delete crlist; PromIS is; // delete list
	  qsort( (void*)buf, kk, sizeof(int), prom_intcompar );
	  ierr = PromISCreateGeneral(PETSC_COMM_SELF,kk,buf,&is);
	  CHKERRQ(ierr);
	  grid->CRAggs_[ii] = is;
	}
	assert(ii==naggs); assert(nn == CC->map_.nlocalEq());
	grid->nCRAggs_ = naggs;
	if( buf != buff__ ) PetscFree(buf);
      }
    }
    else if( flag && strcmp(str,"segregated")==0 ) {
      // make C * C' for Shur smoothers and coarsening (CDinvCt_)
      const PromMatrix *const AA = KK; PromMatrix *CDCt; 
      PromVector *Adiag = new PromVector( AA->map_ );  // scale for CDinvCt
      ierr = AA->GetDiagonal( Adiag );      CHKERRQ(ierr);
      ierr = Adiag->Reciprocal();           CHKERRQ(ierr);
      ierr = Adiag->Sqrt();                 CHKERRQ(ierr);
      PromMatrix *const CT = (PromMatrix *)Ct; // hack past the const
      ierr = CT->Scale( Adiag, NULL );      CHKERRQ(ierr);
      ierr = TransMult( CT, CT, &CC->map_, &CDCt );  CHKERRQ(ierr);
      PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(grid->stiffness_); 
      assert( K->CDinvCt_ == NULL ); K->CDinvCt_ = CDCt;
      ierr = Adiag->Reciprocal();           CHKERRQ(ierr); // scale back
      ierr = CT->Scale( Adiag, NULL );      CHKERRQ(ierr);
      delete Adiag;
    }
    delete tdiag; delete Ti; // done with aggregation matrix
  }
  
  if( 0 ){ // debug
    // make C * C'
    PromMatrix *CCt; 
    const PromMatrix *const Ct=K->Ct_,*const CC=K->C_; assert(Ct!=NULL);
    ierr = TransMult( Ct, Ct, &CC->map_, &CCt );            CHKERRQ(ierr);
    FILE *file;char path[128]; 
    if( npe_ == 1 ) sprintf(path,"%s/CCt_%d",root_,grid->getLevel());
    else sprintf(path,"%s/CCt_%d_%d",root_,grid->getLevel(),mype_); 
    file = fopen(path,"w");  assert(file!=NULL);
    // scale
    PromVector *tdiag = new PromVector( CCt->map_ );
    ierr = CCt->GetDiagonal( tdiag );                      CHKERRQ(ierr);
    ierr = tdiag->Reciprocal();                            CHKERRQ(ierr);
    ierr = tdiag->Sqrt();                                  CHKERRQ(ierr);
    ierr = CCt->Scale( tdiag, tdiag );                     CHKERRQ(ierr);
    // write
    PromContext::WriteMat_private(CCt,file,0,0,false); 
    fclose(file); 
    // check for lin. dependent
    double *Avals; int nA, *Aids; int my0 = CCt->map_.my0Eq();
    for( ii = 0 ; ii < CCt->getm() ; ii++ ){
      ierr = CCt->GetRow( ii+my0, &nA, &Aids, &Avals );   CHKERRQ(ierr);
      for( jj = 0 ; jj < nA ; jj++ ){
	int geqj = Aids[jj]; double vv = fabs(Avals[jj]);
	if(my0+ii == geqj) assert( 1. - vv < 1.e-6 );
	else if(vv>.9)PetscPrintf(PETSC_COMM_SELF,"\t[0]WARNING %d,%d |CCT_ij|=%e (M=%d)\n",my0+ii,geqj,vv,CCt->getM());
      }
      ierr = CCt->RestoreRow( ii+my0, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    }
    delete tdiag; delete CCt;
  }
  if(verbose_>2)PetscPrintf(wcomm.MPIComm(),"\t[0]%s done\n",__FUNCT__);
  ierr = prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_AUX2]); 
  CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#define PROM_MAX_SGAGGS 10000
#define PROM_MAX_AGG_COARSEN_SIZE 100000
#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::AggregateCRs"
/* Prometheus_LinSysCore::AggregateCRs() ***************************
 *
 *  - aggragates CRs, construct P_{cr}
 *
 *  INPUT: 
 *   - aggs: (out)
 *   - Ti: matrix to aggregate on
 *   - tol1:
 *   - tol2:
 *   - do_glob
 *
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::AggregateCRs( PromList<PromAList<int>*> &aggs,
					 const PromMatrix *const Ti, 
					 const double tol2, const bool do_glob,
					 const bool set_with_local/*=false*/ ) const 
{
  int ierr,ii,jj,tt,geqi,nA,leqi,ndone,*Aids,proc,nrecv,wht,maxw,its;
  const int lmLoc=Ti->map_.nlocalEq(),my0Lm=Ti->map_.my0Eq();
  const int *const proc_eq = Ti->map_.proc_geq(), np = Ti->map_.comm_.np();
  double *Avals, vv, sum; 
  const int reqsz = sizeof(MPI_Request)/sizeof(int);
  const int tag=Ti->map_.comm_.getNewTag(),tag1=Ti->map_.comm_.getNewTag();
  const int tag2=Ti->map_.comm_.getNewTag(),tag3=Ti->map_.comm_.getNewTag();
  PromAList<int> *crlist, **lid_aggList, dummy(1); Position pos;
  MPI_Status status;  MPI_Comm wcomm = Ti->map_.comm_();  TablePos tpos; 
  PromList<MPI_Request*> smessages;
  PromTable<PromList<int>*> proc_list(57), proc_list2(57); PromList<int> *list;
  PromTable<int> gid_wht(3*lmLoc + 117); 
  PromTable<PromAList<int>*> ghost_list(3*lmLoc+117), sgid_list(2*lmLoc+17); 
  
  // aggregate (make 'aggs')7
  ierr = PetscMalloc((lmLoc+1)*sizeof(PromAList<int>*),&lid_aggList);
  CHKERRQ(ierr);
  for( ii = 0 ; ii < lmLoc ; ii++ ) lid_aggList[ii] = NULL;
  
  // collect sends, set local wht, create singletons
  for( leqi = ndone = 0, geqi = my0Lm ; leqi < lmLoc ; leqi++, geqi++ ){
    ierr = Ti->GetRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    for(jj=0, sum = 0.0 ; jj < nA ; jj++){
      const int geqj = Aids[jj], leqj = geqj - my0Lm; vv = fabs(Avals[jj]);
      if( leqj == leqi ) assert( (vv-1.0) < 1.e-6 );
      else if ( (leqj < 0 || leqj >= lmLoc) && !do_glob ) /*void*/;
      else if ( vv > tol2 ){
	//sum += vv;
	sum += vv * vv; 
	//if( vv > sum ) sum = vv; 
	if ( leqj < 0 || leqj >= lmLoc ) {
	  PromGrid::FindProc(proc_eq, np, geqj, proc); assert(proc!=mype_);
	  if( (list=proc_list.find(proc+1)) == NULL ) {
	    list = new PromList<int>(16); CHKNEW(list,1);
	    proc_list.add( list, proc+1 );
	  }
	  list->addTail( leqi + 1 );
	}
      }
    }
    ierr = Ti->RestoreRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    //
    if( nA > PROM_MAX_AGG_COARSEN_SIZE || sum == 0.0 ){ // make a singleton
      //PetscPrintf(PETSC_COMM_WORLD,"\t[%d]%s have SING %d/%d %g\n",mype_,__FUNCT__,nA,PROM_MAX_AGG_COARSEN_SIZE,sum);
      wht = -8; // flag for no use
      crlist = new PromAList<int>(5); crlist->count_ = -1;
      aggs.addTail( crlist ); assert( lid_aggList[leqi] == NULL ); 
      lid_aggList[leqi] = crlist; // this is now picked
      crlist->addTail(geqi+1); ndone++; // agg.
      sgid_list.add( crlist, geqi+1 );
    }
    else wht = (int)(1000.0*sqrt(sum)) + 1; 
    //else wht = (int)(1000.0*sum) + 1; 

    gid_wht.add( wht, geqi+1 );
  }
  // send(tag2) *< geqi, wht > -- get ghost whts
  tpos = proc_list.getHeadPosition(); nrecv = 0;
  while( tpos ){
    list = proc_list.getNext( tpos, &proc ); proc--; // skip others procs
    tt = 2 * list->getCount(); 
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq);
    CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    pos = list->getHeadPosition(); 
    while( pos ){
      leqi = list->getNext( pos ) - 1; nrecv++;
      int geqi = leqi + my0Lm; *pb++ = geqi; wht = gid_wht.find(geqi+1);
      *pb++ = wht; assert(wht!=0); 
    }
    if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
    // send(tag2) *< geqi, wht >
    MPI_Isend( buff, tt, MPI_INT, proc, tag2, wcomm, sreq );
  }
  // recv(tag2)  *< geqi, wht >
  while( nrecv > 0 ) {
    MPI_Probe(MPI_ANY_SOURCE, tag2, wcomm, &status );
    MPI_Get_count( &status, MPI_INT, &tt );  assert(tt%2==0);
    proc = status.MPI_SOURCE;     assert(proc != mype_);
    // recieve updates
    int *buff, *pb, __buff[PROM_BSZ];
    if( tt > PROM_BSZ ){
      ierr = PetscMalloc(tt*sizeof(int),&buff );  CHKERRQ(ierr); pb = buff;
    }
    else pb = buff = __buff;
    // recv(tag2) *< geqi, wht > 
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag2, wcomm, &status );
    while( pb-buff < tt ) {
      geqi = *pb++; wht = *pb++; nrecv--; assert(nrecv>=0); 
      gid_wht.add( wht, geqi+1 ); assert(wht!=0); 
    }
    assert(pb-buff == tt); if(buff != __buff) PetscFree(buff); 
  }
  // cleanup sends 
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe( MPI_ANY_SOURCE, tag2, wcomm, &tt, &status );   assert(!tt);
  
  // MIS
  int locAggID = PROM_MAX_SGAGGS * mype_; // processor encoded -- not used
  for( its = 0 ; its < 9 ; its++ ){
    if(its==9-1)PetscPrintf(PETSC_COMM_WORLD,"[%d] Warning: last MIS iter!!!\n",mype_);
    int nd;
    do{
      nd = 0;
      for(leqi=0, geqi=my0Lm ; leqi < lmLoc && ndone < lmLoc ; leqi++, geqi++){
	if( lid_aggList[leqi] != NULL ) continue;
	ierr = Ti->GetRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
	int whti = gid_wht.find(geqi+1); assert(whti>0);
	// see if I am selectable, and collect current wht (sum) and max w
	for( maxw = jj = 0, sum = 0.0 ; jj < nA ; jj++ ){
	  const int geqj = Aids[jj], leqj = geqj - my0Lm; vv = fabs(Avals[jj]);
	  if( leqj == leqi ) assert((vv-1.0)<1.e-6);
	  else if ( (leqj < 0 || leqj >= lmLoc) && !do_glob ) /*void*/;
	  else if ( vv > tol2 ){
	    PromAList<int> *crlistJ; 
	    if( leqj >= 0 && leqj < lmLoc ) crlistJ = lid_aggList[leqj]; 
	    else crlistJ = ghost_list.find(geqj+1);
	    //
	    if( crlistJ == NULL ) { // test
	      wht = gid_wht.find(geqj+1); 
	      if( wht < 0 ) continue;  assert( wht > 0 );
	      //sum += vv;
	      sum += vv * vv; 
	      //if( vv > sum ) sum = vv;
	      // colect max wht
	      if( leqj < 0 || leqj >= lmLoc ) {
		if( wht > maxw ) maxw = wht;
		if( wht == whti ) {
		  PromGrid::FindProc(proc_eq,np,geqj,proc);assert(proc!=mype_);
		  if( proc > mype_ ) maxw = wht + 1; // break ties
		}
	      }
	      else if(wht>maxw) maxw = wht;//the max of wht and of max neigh w
	    }
	  }
	}
	wht = (int)(1000.0*sqrt(sum)) + 1; 
	//wht = (int)(1000.0*sum) + 1; 
	gid_wht.add( wht, geqi+1 ); 
	// doit?
	if( wht >= maxw ) {
	  nd++;  // doit
	  crlist = new PromAList<int>(5); crlist->count_ = -1;
	  aggs.addTail( crlist );
	  lid_aggList[leqi] = crlist; // this is now picked
	  crlist->addTail( geqi + 1 ); ndone++; // agg.
	  sgid_list.add( crlist, geqi+1 ); // selected nodes
	  for( jj = 0 ; jj < nA ; jj++ ){
	    const int geqj=Aids[jj], leqj = geqj-my0Lm; vv=fabs(Avals[jj]);
	    if( leqj == leqi ) assert((vv-1.0)<1.e-6);
	    else if ( (leqj < 0 || leqj >= lmLoc) && !do_glob ) /*void*/;
	    else if ( vv > tol2 ){
	      PromAList<int> *crlistJ;
	      if( leqj >= 0 && leqj < lmLoc ) crlistJ = lid_aggList[leqj]; 
	      else crlistJ = ghost_list.find(geqj+1);
	      //
	      if( crlistJ == NULL ) {	
		if( leqj < 0 || leqj >= lmLoc ){
		  int add = ghost_list.add( crlist, geqj+1 ); assert(add==1);
		  if(crlist->count_ == -1)crlist->count_ = locAggID++; //flag: has a ghost
		  // collect message to send
		  PromGrid::FindProc(proc_eq, np, geqj, proc); assert(proc!=mype_);
		  if( (list=proc_list2.find(proc+1)) == NULL ) {
		    list = new PromList<int>(9); CHKNEW(list,1);
		    proc_list2.add( list, proc+1 );
		  }
		  list = proc_list2.find(proc+1); assert(list!=NULL);
		  // cache < geqj-D, geqi-S >
		  list->addTail( geqj + 1 ); list->addTail( geqi + 1 );
		}
		else{ lid_aggList[leqj] = crlist; ndone++; }// agg.
	      }
	    }
	  }
	}
	ierr = Ti->RestoreRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
      }
    }while( nd > 0 );

    // send(tag) *< geqj-D, geqi-S > 
    tpos = proc_list2.getHeadPosition(); 
    while( tpos ){
      list = proc_list2.getNext( tpos, &proc ); proc--; // skip others procs
      tt = list->getCount(); assert(tt%2==0);
      if( tt > 0 ) {
	MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq);
	CHKERRQ(ierr);
	smessages.addTail( sreq );
	int *buff = (int*)sreq + reqsz, *pb = buff;
	while( !list->isEmpty() ) *pb++ = list->removeHead() - 1;
	if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
	// send(tag) *< geqj-D, geqi-S >
	MPI_Isend( buff, tt, MPI_INT, proc, tag, wcomm, sreq );
      }
    }
    // send(tag1) capper <gidi,wht,[PROM_SELECTED|PROM_DELETED|PROM_UNKNOWN] >
    tpos = proc_list.getHeadPosition(); nrecv = 0;
    while( tpos ){
      list = proc_list.getNext( tpos, &proc ); proc--; // skip others procs
      tt = 3 * list->getCount(); 
      MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq);
      CHKERRQ(ierr);
      smessages.addTail( sreq );
      int *buff = (int*)sreq + reqsz, *pb = buff;
      pos = list->getHeadPosition(); 
      while( pos ){
	leqi = list->getNext( pos ) - 1; nrecv++;
	int geqi = leqi + my0Lm; *pb++ = geqi; 	
	crlist = sgid_list.find( geqi+1 );
	PromNodeState st = (crlist!=NULL) ? PROM_SELECTED : (lid_aggList[leqi]!=NULL) ? PROM_DELETED : PROM_UNKNOWN;
	*pb++ = (int)st;
	*pb++ = gid_wht.find(geqi+1);
      }
      if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
      // send(tag1) *<geqi, wht, [PROM_SELECTED|PROM_DELETED|PROM_UNKNOWN] >
      MPI_Isend( buff, tt, MPI_INT, proc, tag1, wcomm, sreq );
    }
    // recv(tag1)  *<geqi, wht, [PROM_SELECTED|PROM_DELETED|PROM_UNKNOWN] >
    while( nrecv > 0 ) {
      MPI_Probe(MPI_ANY_SOURCE, tag1, wcomm, &status );
      MPI_Get_count( &status, MPI_INT, &tt );  assert(tt%3==0);
      proc = status.MPI_SOURCE;     assert(proc != mype_);
      // recieve updates
      int *buff, *pb, __buff[PROM_BSZ];
      if( tt > PROM_BSZ ){
	ierr = PetscMalloc(tt*sizeof(int),&buff );  CHKERRQ(ierr); pb = buff;
      }
      else pb = buff = __buff;
      // recv(tag1) *<geqi, wht, [PROM_SELECTED|PROM_DELETED|PROM_UNKNOWN] > 
      MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag1, wcomm, &status );
      while( pb-buff < tt ) {
	geqi = *pb++; nrecv--; assert(nrecv>=0); 
	PromNodeState st = (PromNodeState)*pb++;
	if(st==PROM_DELETED && !ghost_list.find(geqi+1))ghost_list.add(&dummy,geqi+1);
	wht = *pb++;
	gid_wht.add( wht, geqi+1 );
      }
      assert(pb-buff == tt); if(buff != __buff) PetscFree(buff); 
    }
    // recv(tag) *< geqi-D, geqj-S >
    while( 1 ){
      MPI_Iprobe( MPI_ANY_SOURCE, tag, wcomm, &tt, &status );
      if( !tt ) break;
      MPI_Get_count( &status, MPI_INT, &tt ); 
      int *buff, *pb, __buff[PROM_BSZ];
      if( tt > PROM_BSZ ) {
	ierr = PetscMalloc( tt*sizeof(int), &buff ); CHKERRQ(ierr); pb = buff;
      }
      else pb = buff = __buff;
      // recv(tag) *< geqi-D, geqj-S >
      MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, wcomm, &status );
      while( pb-buff < tt ){
	geqi = *pb++; int geqj = *pb++; leqi = geqi - my0Lm; 
	if(lid_aggList[leqi]==NULL){lid_aggList[leqi] = &dummy; ndone++;}// local
	sgid_list.add( &dummy, geqj+1 ); // selected nodes
	ghost_list.add( &dummy, geqj+1 );
      }
      assert(pb-buff == tt);  if(buff != __buff) PetscFree(buff); 
    }
    // cleanup sends 
    while( !smessages.isEmpty() ) {
      MPI_Request *sreq = smessages.removeHead();
      MPI_Wait( sreq, &status );    PetscFree( sreq );
    }
    MPI_Iprobe( MPI_ANY_SOURCE, tag, wcomm, &tt, &status );   assert(!tt);

    // test for done
    tt = (ndone==lmLoc) ? 0 : 1; // !done
    MPI_Allreduce( &tt, &ii, 1, MPI_INT, MPI_SUM, wcomm ); // get max size
    if( ii == 0 ) break; // all done
  }
  MPI_Iprobe( MPI_ANY_SOURCE, tag, wcomm, &tt, &status );   assert(!tt);
  MPI_Iprobe( MPI_ANY_SOURCE, tag1, wcomm, &tt, &status ); assert(!tt);

  // set non-MIS nodes to node agg, collect sends
  for( leqi = 0, geqi = my0Lm ; leqi < lmLoc ; leqi++, geqi++ ){
    if( sgid_list.find( geqi+1 ) ) continue;
    ierr = Ti->GetRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    for( jj = 0, ii = -1, sum = 0.0 ; jj < nA ; jj++ ){
      const int geqj = Aids[jj], leqj = geqj - my0Lm;vv=fabs(Avals[jj]);
      if( leqj == leqi ) assert((vv-1.0)<1.e-6);
      else if ( (leqj < 0 || leqj >= lmLoc) && !do_glob ) /*void*/;
      else if ( vv > tol2 ){
	if( sgid_list.find(geqj+1) ){
	  wht = gid_wht.find(geqj+1); assert(wht!=0);
	  if( wht > 0 && vv > sum ){ sum = vv; ii = geqj; }
	}
      }
    }
    if(ii==-1)SETERRQ1(1,"gid %d not found (singleton??)",geqi);
    ierr = Ti->RestoreRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    const int s_geqj = ii, s_leqj = s_geqj - my0Lm;
    if ( s_leqj < 0 || s_leqj >= lmLoc ){ // ghost
      PromGrid::FindProc( proc_eq, np, s_geqj, proc ); assert(proc!=mype_);
      if( (list=proc_list2.find(proc+1)) == NULL ) {
	list = new PromList<int>(9); CHKNEW(list,1);
	proc_list2.add( list, proc+1 );
      }
      // cache < geqi-D, geqj-S >
      list->addTail( geqi + 1 ); list->addTail( s_geqj + 1 );
    }
    else { // local
      crlist = lid_aggList[leqi] = lid_aggList[s_leqj]; assert(crlist != NULL);
      crlist->addTail( geqi + 1 ); // agg.
    }
  }
  assert(ndone==lmLoc);
  // send(tag3) *< geqi-D, geqj-S > 
  tpos = proc_list2.getHeadPosition(); 
  while( tpos ){
    list = proc_list2.getNext( tpos, &proc ); proc--; // skip others procs
    tt = list->getCount(); assert(tt%2==0);
    if( tt > 0 ) {
      MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq);
      CHKERRQ(ierr);
      smessages.addTail( sreq );
      int *buff = (int*)sreq + reqsz, *pb = buff;
      while( !list->isEmpty() ) *pb++ = list->removeHead() - 1;
      if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
      // send(tag3) *< geqi-D, geqj-S >
      MPI_Isend( buff, tt, MPI_INT, proc, tag3, wcomm, sreq );
    }
  }
  // send(tag3) capper
  tpos = proc_list.getHeadPosition(); nrecv = 0;
  while( tpos ){
    list = proc_list.getNext( tpos, &proc ); proc--; // skip others procs
    MPI_Send( NULL, 0, MPI_INT, proc, tag3, wcomm );
  }
  // recv(tag3) *< geqj-D, geqi-S >
  nrecv = proc_list.getCount();
  while( nrecv > 0 ){
    MPI_Probe( MPI_ANY_SOURCE, tag3, wcomm, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); 
    if( tt == 0 ){
      MPI_Recv( NULL, 0, MPI_INT, status.MPI_SOURCE, tag3, wcomm, &status );
      nrecv--;
    }
    else {
      int *buff, *pb, __buff[PROM_BSZ];
      if( tt > PROM_BSZ ) {
	ierr = PetscMalloc( tt*sizeof(int), &buff ); CHKERRQ(ierr); pb = buff;
      }
      else pb = buff = __buff;
      // recv(tag3) *< geqj-D, geqi-S >
      MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag3, wcomm, &status );
      while( pb-buff < tt ) {
	int geqj = *pb++; geqi = *pb++; leqi = geqi - my0Lm; 
	crlist = lid_aggList[leqi]; assert(crlist != NULL);
	crlist->addTail( geqj + 1 ); // agg.
	ghost_list.add( crlist, geqj+1 ); 
	if(crlist->count_==-1) crlist->count_ = locAggID++; //flag: has a ghost
      }
      assert(pb-buff == tt);  if(buff != __buff) PetscFree(buff); 
    }
  }
  // cleanup sends 
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe( MPI_ANY_SOURCE, tag3, wcomm, &tt, &status );   assert(!tt);
  //#if defined(PETSC_USE_BOPT_g)
  // debug
  for( ii = 0, sum = 0.0 ; ii < lmLoc ; ii++ ) {
    leqi = ii; geqi = leqi + my0Lm; crlist = lid_aggList[leqi];
    ierr = Ti->GetRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    double sum2 = 0.0;
    for( jj = 0 ; jj < nA ; jj++ ){
      const int geqj = Aids[jj], leqj = geqj - my0Lm; vv = fabs(Avals[jj]);
      if( leqj != leqi ) {
	PromAList<int> *crlistJ; 
	if( leqj >= 0 && leqj < lmLoc ) crlistJ = lid_aggList[leqj]; 
	else crlistJ = ghost_list.find(geqj+1);
	if( crlistJ != crlist && vv > sum2 ) sum2 = vv; 
      }
    }
    if( sum2 > sum ) sum = sum2; 
    ierr = Ti->RestoreRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
  }
  if( lmLoc > 0 && verbose_ > 1 ){
    PetscPrintf(PETSC_COMM_SELF,"[%d] MIS DONE, %d aggs, max edge cut = %g (%d CRs)\n",
		mype_,aggs.getCount(),sum,lmLoc);
  }
  //#endif
  // clean up
  PetscFree( lid_aggList );
  while( !proc_list.isEmpty() ) delete proc_list.removeHead();
  while( !proc_list2.isEmpty() ) delete proc_list2.removeHead();

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::MoveAggs"
/* Prometheus_LinSysCore::MoveAggs() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *   - map: 
 *   - aggs: aggs have 'count_ == -1' for interior aggs (in/out)
 *   - nextFactor:
 *   - CP_colMap:
 *   - CP:
 *
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::MoveAggs( const PromMap &map,
				     PromList<PromAList<int>*> &aggs,
				     const int nextFactor,
				     const PromMap*const CP_colMap,
				     const PromMatrix*const CP )const 
{
  int ierr,ii,jj,tt,proc,leq,geq,id;
  const int tag =  map.comm_.getNewTag();
  const int *prim_geq = CP_colMap->proc_geq();
  const int *cr_gnode = map.proc_gnode(),np=map.comm_.np();
  const int lmLoc=map.nlocalEq(),my0Lm=map.my0Eq();
  MPI_Comm wcomm = map.comm_(); 
  Position pos2, pos, opos; MPI_Status status;
  const int reqsz = sizeof(MPI_Request)/sizeof(int);
  PromList<MPI_Request*> rmessages; PromAList<int> *crlist;
  int rbuf[PROM_BSZ],sbuf[PROM_BSZ]; 

  // remove lists, collect sends
  PromList<int> tlist;
  pos = aggs.getHeadPosition(); 
  while( (opos=pos) ){
    crlist = aggs.getNext( pos ); // skip others procs
    if( crlist->isEmpty() ) {
      aggs.removeAt( opos );
      delete crlist;
    }
    else {
      assert(crlist->count_ == -1 || crlist->count_/PROM_MAX_SGAGGS==mype_);
      // get highest ID, and collect proc counts
      int nn = 1; sbuf[0] = 0; rbuf[0] = mype_;
      pos2 = crlist->getHeadPosition(); assert(pos2!=NULL);
      while( pos2 ) {
	geq = crlist->getNext( pos2 ) - 1; leq = geq - my0Lm; 
	if( leq < 0 || leq >= lmLoc ) { // ghost
	  PromGrid::FindProc( cr_gnode, np, geq, proc ); int i;
	  proc = (proc/nextFactor)*nextFactor;
	  if( proc != mype_ ) { 
	    for(i=1;i<nn;i++) if(rbuf[i] == proc) break;
	    if(i == nn){ sbuf[nn] = 1; rbuf[nn] = proc; nn++; }
	    else sbuf[i]++;
	  }
	  else{ sbuf[0]++; assert(nextFactor > 1); }
	}
	else{ // local-check for a diff proc-- use CP each column, thus > wieght
	  int nA, *Aids; 
	  ierr = CP->GetRow( geq, &nA, &Aids );   CHKERRQ(ierr);
	  for( jj = 0 ; jj < nA ; jj++ ) {
	    int pgeq = Aids[jj], i;
	    PromGrid::FindProc( prim_geq, np, pgeq, proc );
	    assert(proc%nextFactor == 0);
	    if( proc == mype_ ) sbuf[0]++; // typical case
	    else{
	      for(i=1;i<nn;i++) if(rbuf[i] == proc) break;
	      if(i == nn){ sbuf[nn] = 1; rbuf[nn] = proc; nn++; }
	      else sbuf[i]++;
	    }
	  }
	  ierr = CP->RestoreRow( geq, &nA, &Aids ); CHKERRQ(ierr);	
	}
      }
      
      // my proc, i'm highest, now put on best
      jj = 0;  proc = -1;
      for( int i = 0 ; i < nn ; i++ ) {
	if( sbuf[i] > jj ){ jj = sbuf[i]; proc =  rbuf[i]; }
      }
      assert(proc != -1 && proc%nextFactor == 0);
      proc = (proc/nextFactor) * nextFactor; // avoid the reduced procs
      if( proc != mype_ ) { // move it
	// cache < proc, ID, n,  n*<geq> >
	tlist.addTail( proc+1 ); 
	tt = PROM_MAX_SGAGGS*proc + PROM_MAX_SGAGGS-1; // dummy id on proc
	tlist.addTail( tt + 1 ); // move a local
	tlist.addTail( crlist->getCount() );
	while( !crlist->isEmpty() ) tlist.addTail( crlist->removeHead() );
	aggs.removeAt( opos );
	delete crlist;
      }
    }
  }
  // count sends
  for( ii = 0 ; ii < np ; ii++ ) sbuf[ii] = 0; 
  pos = tlist.getHeadPosition(); tt = 0;
  while( pos ){
    // cache < proc, ID, n,  n*<geq> >
    proc = tlist.getNext( pos ) - 1; ii = tlist.getNext( pos ) - 1;
    sbuf[proc]++; ii = tlist.getNext( pos );
    if( ii > tt ) tt = ii;
    while( ii-- ) tlist.getNext( pos );
  }
  // global!
  MPI_Alltoall(sbuf,1,MPI_INT,rbuf,1,MPI_INT,wcomm);
  MPI_Allreduce( &tt, &ii, 1, MPI_INT, MPI_MAX, wcomm ); // get max size
  const int rsz = ii+1;
  // post recvs
  for(proc=0;proc<np;proc++) {
    while( rbuf[proc]-- > 0 ) {
      int *pb; MPI_Request *rreq;
      ierr = PetscMalloc((rsz+reqsz)*sizeof(int), &rreq ); CHKERRQ(ierr);
      pb = (int*)(rreq+1);
      // irecv(tag) < ID, *<geq> >
      MPI_Irecv( pb, rsz, MPI_INT, proc, tag, wcomm, rreq );
      rmessages.addTail( rreq );
    }
  }
  // send
  pos = tlist.getHeadPosition(); 
  while( pos ){
    // cache < proc, ID, n, n*<geq> >
    proc = tlist.getNext(pos) - 1; id = tlist.getNext(pos) - 1;
    ii = 0; sbuf[ii++] = id; 
    tt = tlist.getNext(pos);
    while( tt-- ) sbuf[ii++] = tlist.getNext( pos ) - 1;
    // send(tag) < ID, *<geq> >
    MPI_Send( sbuf, ii, MPI_INT, proc, tag, wcomm );
  }

  // recv(tag) new lists
  while( !rmessages.isEmpty() ) {
    MPI_Request *req = rmessages.removeHead(); 
    // recv(tag) < ID, *<geq> >
    MPI_Wait( req, &status ); 
    int *buff = (int*)(req + 1), *pb = buff;   
    MPI_Get_count( &status, MPI_INT, &tt );
    id = *pb++; proc = id/PROM_MAX_SGAGGS; assert(proc==mype_);
    crlist = NULL; 
    // new list
    crlist = new PromAList<int>(tt); crlist->count_ = id;
    aggs.addTail( crlist );
    // add to list
    while( pb-buff < tt ) crlist->addTail( (*pb++) + 1 );

    if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
    PetscFree(req); 
  }
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::MakeCRProlMap"
/* Prometheus_LinSysCore::MakeCRProlMap() ***************************
 *
 *  - creates 'prol_D_out'
 *
 *  INPUT: 
 *   - aggs: (emptied for data) has 'count_ == -1' for locals
 *   - grid:
 *   - clm_out: this map is used to create C_out
 *   - prol_D_out: (out) 
 *
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::MakeCRProlMap(PromList<PromAList<int>*>&aggs,
					 const PromGrid * const grid, 
					 PromMap **clm_out,
					 PromMatrix **prol_D_out ) const
{
  int ierr,ii,eq,nccrs;
  PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(grid->stiffness_);
  const PromMap &flmap_i = K->C_->map_;
  const PromGrid *const cgrid = grid->nextgrid_;
  const int lmLoc = flmap_i.nlocalEq(), my0Lm = flmap_i.my0Eq();
  Position pos,pos2,pos3; const PromComm &comm = prom_->Comm_;
  int *nnz,*Onnz;  PromAList<int> *crlist;

  // make coarse CR map
  PetscFunctionBegin;
  nccrs = 0; // count locals
  pos = aggs.getHeadPosition(); 
  while( pos ){
    crlist = aggs.getNext( pos );
    if(crlist->count_ == -1 || crlist->count_/PROM_MAX_SGAGGS == mype_) nccrs++;
    else PetscPrintf(PETSC_COMM_SELF,"\t[%d]%s have list ID %d?\n",mype_,__FUNCT__,crlist->count_);
  }

  assert( mype_%cgrid->factor_ == 0 || nccrs == 0 ); 
  PromMap *clmap = new PromMap( nccrs, 1, comm );
  *clm_out = clmap; // ret

  // make Pc
  ierr = PetscMalloc(2*(lmLoc+1)*sizeof(int),&nnz);  CHKERRQ(ierr);
  Onnz = nnz + lmLoc;
  for(ii=0;ii<lmLoc;ii++)Onnz[ii]=nnz[ii]=0;

  // create new P for CRs, delete off aggs
  pos = aggs.getHeadPosition(); ii = 0;
  while( (pos2=pos) ){
    crlist = aggs.getNext( pos );
    if( crlist->count_ == -1 || crlist->count_/PROM_MAX_SGAGGS == mype_ ) {
      // for each fine cr
      pos3 = crlist->getHeadPosition(); assert(pos3!=NULL);
      while( pos3 ) {
	int geq = crlist->getNext(pos3) - 1, leq = geq - my0Lm; 
	if(leq >= 0 && leq<lmLoc) nnz[leq]++;
	// else someone elses onnz
	ii++;
      }
    }
    else { // off processor aggregate - use for NZs
      pos3 = crlist->getHeadPosition(); assert(pos3!=NULL);
      while( pos3 ) {
	int geq = crlist->getNext(pos3) - 1, leq = geq - my0Lm; 
	if(leq >= 0 && leq<lmLoc) Onnz[leq]++;
      }
      delete crlist;
      aggs.removeAt( pos2 ); 
    }
  }

  // create prolongator_->D_
  const int ncLmLoc = clmap->nlocal(), ncLmGlb = clmap->nglobal();
  PromMatrix *prol = new PromMatrix(flmap_i, 0, nnz, 0, Onnz, ncLmLoc,ncLmGlb);
  *prol_D_out = prol; // ret
  PetscFree(nnz); //PetscFree(Offnnz);

  // add in to prolongator_->D_ and destroy aggs
  for( eq = clmap->my0Eq() ; !aggs.isEmpty() ; eq++ ){
    crlist = aggs.removeHead();
    int ids[4096];  assert(crlist->getCount()<4096);
    //#define KKT_PROL_FACT(k) (1./(double)k)
#define KKT_PROL_FACT(k) (1./sqrt((double)k))
    double vals[4096], scale = KKT_PROL_FACT( crlist->getCount() );
    for( ii = 0 ; !crlist->isEmpty() ; ii++ )  {
      int geq = crlist->removeHead() - 1; 
      ids[ii] = geq; vals[ii] = scale; 
    }
    assert(ii>0);     delete crlist;
    ierr = prol->SetValues( ii, ids, 1, &eq, vals, INSERT_VALUES );
    CHKERRQ(ierr);
  }
  ierr = prol->Assembly(); CHKERRQ(ierr);
  
  // output Pc
  if( verbose_ > 2 ) {
    PromMatrix *prol = *prol_D_out; int lev = cgrid->getLevel();
    FILE *file; char path[128]; PromMatrix *P = cgrid->prolongator_;
    PetscPrintf(PETSC_COMM_WORLD,"\t[%d]%s output prolongator matrix\n",
		mype_,__FUNCT__);
    if( npe_ == 1 ) sprintf( path, "%s/D_%d", root_, lev );
    else sprintf( path, "%s/D_%d_%d", root_, lev, mype_);
    file = fopen( path, "w");
    ierr = PromContext::WriteMat_private(prol,file,P->getM(),P->getN(),false);
    CHKERRQ(ierr);
    fclose(file);
    if( npe_ == 1 ) sprintf( path, "%s/P_%d", root_, lev );
    else sprintf( path, "%s/P_%d_%d", root_, lev, mype_);
    file = fopen( path, "w");
    ierr = PromContext::WriteMat_private(P,file,0,0,false);CHKERRQ(ierr);
    fclose(file);
  }

  PetscFunctionReturn(0);
}

class prom_int_double {
public:
  prom_int_double(int cid=0, int rid=0, double v=0.0) : 
    colid_(cid), rowid_(rid), val_(v){}
  int colid_;
  int rowid_;
  double val_;
};


#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::Mult"
/* Prometheus_LinSysCore::Mult() ***************************
 *
 *  - creates 'C_out', <-- AA * BB 
 *
 *  INPUT: 
 *   - AA: row orient (map is used to create C_out)
 *   - BB: row orient
 *   - C_out: (out) uses 'AA' map.
 *
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::Mult( const PromMatrix * const AA, 
				 const PromMatrix *const BB, 
				 PromMatrix **C_out )
{
  PromMatrix *AB = NULL; assert(!AA->isComplex());
  double *Avals, *Bvals, Aik, *fp; const int tag = BB->map_.comm_.getNewTag();
  const int A0eq=AA->map_.my0Eq(), nAloc=AA->map_.nlocalEq();
  const int B0eq=BB->map_.my0Eq(), nxtB=B0eq+BB->map_.nlocalEq();
  const int *const proc_beq = BB->map_.proc_geq(), np = BB->map_.comm_.np();
  const int rsz = 2 + sizeof(double)/sizeof(int), mype=BB->map_.comm_.mype();
  const int reqsz = sizeof(MPI_Request)/sizeof(int);
  MPI_Status status;
  MPI_Comm wcomm = BB->map_.comm_();  
  PromList<MPI_Request*> smessages;
  PromTable<PromList<prom_int_double*>*> proc_list(57);
  PromList<prom_int_double*> *list; 
  int bfsz,ierr,ii,kk,jj,*Aids,*Bids,nA,nB,II,proc,gidj,tt;
  double *fbuf, fbuff__[4096]; bfsz = 4096; fbuf = fbuff__;
  
  PetscFunctionBegin;
  { 
    int m=AA->getm()+1,onz=15,nnz=2*(AA->getLocalNNZ()+BB->getLocalNNZ())/m;
    if( nnz > BB->getn() ) nnz = BB->getn(); 
    if( onz > BB->getN()-BB->getn() ) onz = BB->getN()-BB->getn(); 
    do{
      AB = new PromMatrix(AA->map_, nnz,NULL, onz,NULL, BB->getn(),BB->getN());
      if( AB && !AB->isOK() ){ delete AB; AB = NULL; }
      nnz /= 2; onz /= 2;
    }while( AB == NULL && nnz > 0 );
  }
  //
  for( ii = 0, II = A0eq ; ii < nAloc ; ii++, II++ ){
    ierr = AA->GetRow( II, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    for( kk = 0 ; kk < nA ; kk++ ){
      gidj = Aids[kk]; Aik = Avals[kk];
      if( gidj >= B0eq && gidj < nxtB ) {
	ierr = BB->GetRow( gidj, &nB, &Bids, &Bvals );   CHKERRQ(ierr);
	if( nB > bfsz ) { 
	  if( fbuf != fbuff__ ) PetscFree(fbuf);
	  bfsz = nB;
	  ierr = PetscMalloc(bfsz*sizeof(double),&fbuf); CHKERRQ(ierr);
	}
	for( jj = 0 ; jj < nB ; jj++ ) fbuf[jj] = Bvals[jj] * Aik;
	// set
	ierr = AB->SetValues(1, &II, nB, Bids, fbuf,ADD_VALUES);CHKERRQ(ierr);
	ierr = BB->RestoreRow(gidj, &nB, &Bids, &Bvals ); CHKERRQ(ierr);
      }
      else {
	PromGrid::FindProc( proc_beq, np, gidj, proc ); assert(proc!=mype);
	// send request for scaled gidj row: <gidj, II, Aik>
	if( (list=proc_list.find(proc+1)) == NULL ) {
	  list = new PromList<prom_int_double*>(16); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	}
	prom_int_double *crv = new  prom_int_double(gidj, II, Aik);
	list->addTail( crv );
      }
    }
    ierr = AA->RestoreRow( II, &nA, &Aids, &Avals ); CHKERRQ(ierr);
  }
  { // send reqests -- out: 'tt'
    int rbuf[PROM_BSZ], sbuf[PROM_BSZ];
    for( ii = 0 ; ii < np ; ii++ ) sbuf[ii] = 0;
    while( !proc_list.isEmpty() ) {
      list = proc_list.removeHead( &proc ); proc--;
      tt = list->getCount() * rsz; assert(tt>0);
      MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq);
      CHKERRQ(ierr);
      smessages.addTail( sreq );
      int *buff = (int*)sreq + reqsz, *pb = buff;
      while( !list->isEmpty() ){ 
	prom_int_double *crv = list->removeHead(); 
	*pb++ = crv->colid_; *pb++ = crv->rowid_; 
	fp = (double*)pb; *fp++ = crv->val_; pb = (int*)fp;
	delete crv;
      }
      if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
      // send(tag) *< rowb, rowC, Aij >
      MPI_Isend( buff, tt, MPI_INT, proc, tag, wcomm, sreq );
      delete list;
      sbuf[proc] = 1;
    }
    for(ii=0;ii<np;ii++) rbuf[ii] = 1;
    // global!
    MPI_Reduce_scatter( sbuf, &tt, rbuf, MPI_INT, MPI_SUM, wcomm );
  }
  // recv
  int nrecvs = tt; 
  while( nrecvs-- > 0 ) {
    MPI_Probe(MPI_ANY_SOURCE, tag, wcomm, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); 
    proc = status.MPI_SOURCE; 
    assert(proc != mype);
    // recieve updates
    int *buff, *pb;
    ierr = PetscMalloc(tt*sizeof(int),&buff );  CHKERRQ(ierr);    pb = buff;
    // recv(tag) *< rowb, rowC, Aij >
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, wcomm, &status );
    while( pb-buff < tt ) {
      gidj = *pb++; II = *pb++; fp = (double*)pb; Aik = *fp++; pb = (int*)fp;
      assert( gidj >= B0eq && gidj < nxtB );
      // get row and compute Aij * and add
      ierr = BB->GetRow( gidj, &nB, &Bids, &Bvals );   CHKERRQ(ierr);
      if( nB > bfsz ) { 
	if( fbuf != fbuff__ ) PetscFree(fbuf);
	bfsz = nB;
	ierr = PetscMalloc(bfsz*sizeof(double),&fbuf); CHKERRQ(ierr);
      }
      for( jj = 0 ; jj < nB ; jj++ ) fbuf[jj] = Bvals[jj] * Aik;
      // set -- communication
      ierr = AB->SetValues(1, &II, nB, Bids, fbuf, ADD_VALUES); CHKERRQ(ierr);
      //
      ierr = BB->RestoreRow( gidj, &nB, &Bids, &Bvals ); CHKERRQ(ierr);
    }
    PetscFree(buff); assert(pb-buff == tt);
  }

  if( fbuf != fbuff__ ) PetscFree(fbuf);
  ierr = AB->Assembly();   CHKERRQ(ierr);
  *C_out = AB; // ret

  // debug
  MPI_Iprobe(MPI_ANY_SOURCE, tag, wcomm, &tt, &status );     assert(!tt);

  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::TransMult"
/* Prometheus_LinSysCore::TransMult() ***************************
 *
 *  - C_out['] <- AA' * BB
 *
 *  INPUT: 
 *   - AA: row orient matrix (column orient operator, AA')
 *   - BB: row orient
 *   - map: 
 *   - C_out: (out)
 *   - Ct_out: (out)
 *
 *  SIDE EFFECTS:
 */ 
int Prometheus_LinSysCore::TransMult(const PromMatrix * const AA, 
				     const PromMatrix * const BB,
				     const PromMap * const map, 
				     PromMatrix **C_out,
				     const PromMap * const mapt /*=NULL*/, 
				     PromMatrix **Ct_out /*=NULL*/ )
{
  PromMatrix *AtB = NULL, *AtBT = NULL; assert(!AA->isComplex());
  int bfsz,ierr,ii,kk,jj,nn,*Aids,*Bids,nA,nB,kkk,ibuf__[4096], *ibuf=ibuf__;
  const int A0eq=AA->map_.my0Eq(),nAloc=AA->getm();
  double *Avals, *Bvals, *fbuf, fbuff__[4096]; bfsz = 4096; fbuf = fbuff__;
  const int nbG=BB->getGlobalNNZ(),naG=AA->getGlobalNNZ(),nBN=BB->getN();
  const int nBn=BB->getn(),nAN=AA->getN(),nAn=AA->getn();
  const int nbL=BB->getLocalNNZ(),naL=AA->getLocalNNZ();

  PetscFunctionBegin;
  if( C_out != NULL ) {
    int nnz=5+20*(naG+nbG)/(nAN+1)+30*(naL+nbL)/(nAn+1),onnz=nnz;
    if(nnz<500)nnz=500;
    if( nnz > nBn ) nnz = nBn;
    if( onnz > nBN-nBn ) onnz = nBN-nBn;
    do{
      AtB = new PromMatrix(*map, nnz, NULL, onnz, NULL, nBn,nBN);
      if( AtB && !AtB->isOK() ){ delete AtB; AtB = NULL; }
      nnz /= 2; onnz /= 2;
    }while( AtB == NULL && nnz > 0 );
    //ierr = AtB->SetOption(MAT_DO_NOT_USE_INODES); CHKERRQ(ierr);
  }
  if( Ct_out != NULL ) {
    int nnz=5+20*(naG+nbG)/(nBN+1)+30*(naL+nbL)/(nBn+1),onnz=nnz;
    if(nnz<500)nnz=500;
    if( nnz > nAn ) nnz = nAn;
    if( onnz > nAN-nAn ) onnz = nAN-nAn;
    do{
      AtBT = new PromMatrix(*mapt, nnz, NULL,onnz,NULL,nAn,nAN);
      if( AtBT && !AtBT->isOK() ){ delete AtBT; AtBT = NULL; }
      nnz /= 2; onnz /= 2;
    }while( AtBT == NULL && nnz > 0 );
    //ierr = AtBT->SetOption(MAT_DO_NOT_USE_INODES); CHKERRQ(ierr);
  }
  for( kk = 0, kkk = A0eq ; kk < nAloc ; kk++, kkk++ ){
    ierr = AA->GetRow( kkk, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    //
    if( BB == AA ) { nB = nA; Bids = Aids; Bvals = Avals; }
    else {
      ierr = BB->GetRow( kkk, &nB, &Bids, &Bvals );   CHKERRQ(ierr);
    }
    if( nB > bfsz ) { 
      if( fbuf != fbuff__ ) PetscFree(fbuf);
      bfsz = nB;
      ierr=PetscMalloc(bfsz*(sizeof(double)+sizeof(int)),&fbuf);CHKERRQ(ierr);
      ibuf = (int*)(fbuf + bfsz);
    }
    for( ii = 0 ; ii < nA ; ii++ ){
      int II = Aids[ii]; double Aik = Avals[ii], temp;
      for( nn = jj = 0 ; jj < nB ; jj++ ) {
	temp = Bvals[jj] * Aik;
	fbuf[nn] = temp; ibuf[nn] = Bids[jj]; nn++;
      }
      if( nn > 0 ) {
	if( AtBT != NULL ) {
	  ierr = AtBT->SetValues(nn,ibuf,1,&II,fbuf,ADD_VALUES);CHKERRQ(ierr);
	}
	if( AtB != NULL )  {
	  ierr = AtB->SetValues(1,&II,nn,ibuf,fbuf,ADD_VALUES);CHKERRQ(ierr);
	}
      }
    }
    if( BB != AA ) {
      ierr = BB->RestoreRow( kkk, &nB, &Bids, &Bvals ); CHKERRQ(ierr);
    }
    ierr = AA->RestoreRow( kkk, &nA, &Aids, &Avals ); CHKERRQ(ierr);
  }
  if( fbuf != fbuff__ ) PetscFree(fbuf);
  //
  if(AtB!=NULL){ierr = AtB->Assembly(); CHKERRQ(ierr); *C_out = AtB;}
  if(AtBT!=NULL){ierr = AtBT->Assembly();CHKERRQ(ierr);*Ct_out= AtBT;}
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::RemoveCRfromAllGrids"
/* Prometheus_LinSysCore::RemoveCRfromAllGrids() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::RemoveCRfromAllGrids() 
{
  PRINT;
  assert(!use_uzawa_); assert(CR_id_->isEmpty());

  // delete CR part of smoothers smoothers
  PromMG *mg; PromGrid *grid, *nextg; PromPC_base *pc = NULL;
  for( grid = prom_->grid0_ ; grid ; grid = nextg ) {
    PromMatrix *KK = grid->stiffness_;  
    PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK); assert(K && K->C_!=NULL); 
    delete K->C_;        K->C_ = NULL;        assert(grid->CRmap_ != NULL);
    delete K->Ct_;        K->Ct_ = NULL;  
    delete grid->CRmap_;  grid->CRmap_ = NULL;  
    if( K->CDinvCt_ != NULL ) delete K->CDinvCt_;  K->CDinvCt_ = NULL; 
    nextg = grid->nextgrid_;
    if( nextg != NULL ) {
      PromCRMatrix *PP = dynamic_cast<PromCRMatrix*>(nextg->prolongator_);
      assert(PP != NULL && PP->D_ != NULL); 
      delete PP->D_; PP->D_ = NULL;
    }
    if( grid == prom_->grid0_ ) {
      pc = prom_->solver_->pc_;
      PromPCMG *mgpc = dynamic_cast<PromPCMG*>(pc);
      if( mgpc != NULL ) {
	// make matrices -- this is done on grid setup so must be done here!
	mg = mgpc->mgstack_; // int loop 
	pc = mg->presmooth_->pc_; 
      }
      else assert(nextg == NULL);
    }
    else {
      mg = mg->next_; // inc loop
      pc = mg->presmooth_->pc_; // cmg?
    }
    assert( pc != NULL && pc->kkt_ != NULL );
    delete pc->kkt_; pc->kkt_ = NULL;
  }
  assert(C_ == NULL);

  lmap_ = NULL; // deleted in 'grid->CRmap_'
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::putCoords_private"
/* Prometheus_LinSysCore::putCoords_private() ***************************
 *
 *  - set coordinate data
 *
 *  INPUT: 
 *
 *
 *   - data: coords[3*numNodes] x1,y1,z1,x2,...
 *
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::putCoords_private( int fieldID, 
					      int fieldSize,
					      int *nodeNumbers, 
					      int numNodes,
					      const double* data,
					      PromVector *right /*=NULL*/ )
{
  PRINT3(fieldID,fieldSize,numNodes);
  PromGrid *grid = prom_->grid0_;
  if( grid == NULL ) SETERRQ( 1, "grid0 == NULL" ); 
  int ierr,ii,jj,kk,gid,lid,xx,tt,proc,todo; ii=sizeof(int); xx=sizeof(double);
  const int fisz = xx/ii,tag=grid->getNewTag();
  jj = sizeof(MPI_Request);
  const int reqsz=(jj/xx + ((jj%xx == 0) ? 0 : 1))*fisz,my0=proc_gnode_[mype_];
  const int nnodes = grid->nNodes_, nloc = grid->nLocalNd_;
  double *coords = NULL;  
  MPI_Status status; 
  PromGlobalID        *dep; 
  PromNode *curr, *const*const node_arr = grid->nodes_;
  PromList<PromNode*> *list; 
  PromList<MPI_Request*> smessages;
  PromTable<PromList<PromNode*>*> proc_list(57);

  if( data != NULL ) {
    if( mype_==0 && verbose_ > 3 && nodeNumbers != NULL ) {
      PromContext::printArr(stderr,nodeNumbers,numNodes,mype_,
			    "nodeNumbers[%d] mype=%d");
      PromContext::printArr(stderr,data,3*numNodes,"data[%d]",fieldSize);
    }
    
    if(fieldSize!=3){
      SETERRQ3(1,"fieldSize != 3 (fieldSize=%d) fieldID=%d numNodes=%d",
	       fieldSize,fieldID,numNodes);
    }
    if( numNodes < nloc ){
      if(verbose_>1)
	PetscPrintf(PETSC_COMM_SELF,
		    "\t\t[%d]Prometheus Warning: NumNodes(%d)<nLocal(%d)\n",
		    mype_, numNodes, nloc );
    }
    if( nnodes > 0 ) {
      jj = 3 * nnodes;
      ierr = PetscMalloc(jj*sizeof(double),&coords);CHKERRQ(ierr);
      while( jj-- ) coords[jj] = -999.;
    }
    // clear flags
    for( xx = 0 ; xx < nnodes ; xx++ ) node_arr[xx]->mark(false);
    // collect nodes
    for( xx = kk = 0 ; xx < numNodes ; xx++, kk += 3 ){
      gid = (nodeNumbers==NULL) ? my0 + xx : nodeNumbers[xx]; lid = gid - my0; 
      if( lid < 0 || lid >= nloc ) {  // ghost
	curr = grid->ghosts_.find(gid+1);
	if( curr == NULL ) continue;
	lid = curr->index_();
      }
      curr = grid->nodes_[lid];
      if( curr->index_.global() != gid ) {
	SETERRQ4(1,"%d/%d) index.global(%d) != gid=%d %d/%d",
		 xx,nloc,curr->index_.global(),gid);
      }
      // collect for send
      if( curr->isExtended() && !curr->isGhost() && !curr->isMarked() ) {
	curr->mark(); // send once (to many procs)
	Position pos = curr->symbData_->dependents->getHeadPosition(); 
	while( pos ) { // send to old ghosts
	  dep = curr->symbData_->dependents->getNext( pos );   
	  proc = dep->proc();
	  if( (list=proc_list.find(proc+1)) == NULL ) {
	    list = new PromList<PromNode*>(104); CHKNEW(list,1);
	    proc_list.add( list, proc+1 );
	  }
	  list->addTail( curr ); 
	}
      }
      // copy data in
      for(ii=0,jj=3*lid;ii<3;ii++,jj++)coords[jj]=data[kk+ii];
    }
    // send
    while( !proc_list.isEmpty() ) {
      list = proc_list.removeHead( &proc ); proc--;
      tt = 4*fisz*list->getCount(); 
      MPI_Request *sreq; ierr = PetscMalloc( (tt+reqsz)*sizeof(int), &sreq ); 
      CHKERRQ(ierr);
      smessages.addTail( sreq );
      int *buff = (int*)sreq + reqsz, *pb = buff;
      assert((size_t)pb%sizeof(double) == 0);
      while( !list->isEmpty() ) {
	curr = list->removeHead();  gid = curr->index_.global(); lid = gid-my0;
	curr->mark(false); // send once
	if( lid < 0 || lid >= nloc ) {
	  SETERRQ4(1,"%d(%d)/%d gid=%d",lid,curr->index_(),nloc,gid);
	}
	*pb++ = gid;
	if( fisz > 1 ) *pb++ = -2;
	double *fb = (double*)pb;
	for( ii = 0, jj = 3*lid ; ii < 3 ; ii++, jj++ ) *fb++ = coords[jj];
	pb = (int*)fb;
      }
      if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
      // send(tag) *< coords >
      MPI_Isend( buff, tt, MPI_INT, proc, tag, grid->gComm_, sreq );
      delete list; 
    }
    // recv
    todo = nnodes - nloc;
    while( todo ) {
      MPI_Probe( MPI_ANY_SOURCE, tag, grid->gComm_, &status );
      MPI_Get_count( &status, MPI_INT, &tt ); proc = status.MPI_SOURCE;
      int *buff; ierr = PetscMalloc(tt*sizeof(int), &buff); int *pb = buff;
      CHKERRQ(ierr);
      // recv(tag) *< coords >
      MPI_Recv(buff, tt, MPI_INT, status.MPI_SOURCE, tag, grid->gComm_,&status);
      while( pb-buff < tt ) {
	todo--; 
	int gid = *pb++; 
	curr = grid->ghosts_.find(gid+1);
	if(curr==NULL)SETERRQ1(1,"curr==NULL: gid=%d",gid);
	if( fisz > 1 ) ii = *pb++;
	double *pv = (double*)pb;
	for( ii=0, jj=3*curr->index_() ; ii<3 ; ii++, jj++ ) {
	  double t = *pv++;
	  if( coords[jj] != -999. && coords[jj] != t ) {
	    PetscPrintf(PETSC_COMM_SELF,"\t[%d]%s ERROR gid %d.%d, %g != %g\n",
			mype_,__FUNCT__,gid,ii,coords[jj],t);
	  }
	  coords[jj] = t;
	}
	pb = (int*)pv;
      }
      if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
      PetscFree(buff); 
    }
  }
  
  // doit
  ierr = grid->SetCoords( coords ); CHKERRQ(ierr);
  assert(!set_coords_); set_coords_ = TRUE;
  if( coords != NULL ) { ierr = PetscFree( coords ); CHKERRQ(ierr); }
  // free
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }

  return 0;
}

/************************************************************************
 * Alternative CR Aggregation
 ************************************************************************/
#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::AggregateCRs2"
/* Prometheus_LinSysCore::AggregateCRs2() ***************************
 *
 *  - aggragates CRs, construct P_{cr}
 *
 *  INPUT: 
 *   - aggs: (out)
 *   - Ti: matrix to aggregate on
 *   - tol1:
 *   - tol2:
 *   - do_glob
 *
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::AggregateCRs2( PromList<PromAList<int>*> &aggs,
					  const PromMatrix *const Ti,
					  const double tol1, const double tol2,
					  const bool do_glob,
					  const bool set_with_local/*=false*/ ) const 
{
  int ierr,ii,jj,geqi,nA,leqi,ndone,*nadj,*Aids;
  const int lmLoc=Ti->map_.nlocalEq(),my0Lm=Ti->map_.my0Eq();
  double *Avals, vv, maxv; 
  PromAList<int> *crlist, **lid_aggList, dummy(1); Position pos;
  PromTable<PromAList<int>*> ghost_list;
  PromTable<PromAList<int>*> listID_list(17);
  
  // aggregate (make 'aggs')
  ierr = PetscMalloc( (2*lmLoc+1)*sizeof(int), &nadj ); CHKERRQ(ierr);
  ierr = PetscMalloc((lmLoc+1)*sizeof(PromAList<int>*),&lid_aggList);
  CHKERRQ(ierr);
  for( ii = 0 ; ii < lmLoc ; ii++ ) lid_aggList[ii] = NULL;

  // strong aggs
  int locAggID = PROM_MAX_SGAGGS * mype_; // processor encoded
  for(leqi = ndone = 0, geqi = my0Lm ; leqi < lmLoc ; leqi++, geqi++ ) {
    ierr = Ti->GetRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    nadj[2*leqi+1] = leqi;
    if( nA > PROM_MAX_AGG_COARSEN_SIZE ) { // make a big singleton
      nadj[2*leqi] = -1; // flag for no use
      crlist = new PromAList<int>(5); crlist->count_ = -1;
      aggs.addTail( crlist ); 
      assert( lid_aggList[leqi] == NULL ); // todo!
      lid_aggList[leqi] = crlist; // this is now picked
      crlist->addTail(geqi+1); ndone++; // agg.
    }  
    else{ 
      //nadj[2*leqi] = nA;
      // MIS with threshhold, get degree
      for(jj=0, crlist = lid_aggList[leqi], maxv=0.0 ; jj < nA ; jj++){
	const int geqj = Aids[jj], leqj = geqj - my0Lm; vv = fabs(Avals[jj]);
	if( leqj == leqi ) assert((vv-1.0)<1.e-6);
	else if ( (leqj < 0 || leqj >= lmLoc) && !do_glob ) /*void*/;
	else {
	  if( vv >= tol1 ) { // 1st threshold
	    PromAList<int> *crlistJ;
	    if( leqj >= 0 && leqj < lmLoc ) crlistJ = lid_aggList[leqj]; 
	    else crlistJ = ghost_list.find( geqj + 1 );
	    // get 'crlist'
	    if( crlist == NULL ) {   // found a string connection - make an agg
	      if( crlistJ != NULL ) crlist = crlistJ; // found another strong agg - add it
	      else { // crlistJ == NULL && crlist == NULL - new list
		crlist = new PromAList<int>(5); crlist->count_ = -1;
		aggs.addTail( crlist ); 
	      }
	      assert(lid_aggList[leqi] == NULL);
	      lid_aggList[leqi] = crlist; crlist->addTail(geqi+1); ndone++;// agg.
	    }
	    // set 'crlistJ'
	    if( crlistJ == NULL ) {
	      if( leqj < 0 || leqj >= lmLoc ){
		int add = ghost_list.add( crlist, geqj+1 ); assert(do_glob&&add==1);
		if( crlist->count_ == -1 ){
		  crlist->count_ = locAggID++; // flag: has a ghost
		  listID_list.add( crlist, crlist->count_ + 1 );
		}
	      }
	      else{
		assert( lid_aggList[leqj] == NULL );
		lid_aggList[leqj] = crlist; ndone++; // agg.
	      }
	      crlist->addTail(geqj+1); 
	    }
	    else if( crlistJ != crlist ){ // convert other 'idj' to 'idi'
	      //PetscPrintf(PETSC_COMM_SELF,"\t\t[%d]%s %d) convert other 'idj' to 'idi'. %d+%d->%d\n",mype_,__FUNCT__,leqi,crlist->getCount(),crlistJ->getCount(),crlist->getCount()+crlistJ->getCount());
	      while( !crlistJ->isEmpty() ) {
		int geqj2 = crlistJ->removeHead() - 1, leqj2 = geqj2 - my0Lm; 
		if( leqj2 < 0 || leqj2 >= lmLoc ){ // a ghost
		  ghost_list.add( crlist, geqj2 + 1 ); assert(do_glob);
		  if( crlist->count_ == -1 ){
		    crlist->count_ = locAggID++; //flag: has a ghost
		    listID_list.add( crlist, crlist->count_ + 1 );
		  }
		}
		else lid_aggList[leqj2] = crlist;
		crlist->addTail( geqj2 + 1 ); // agg. (allready counted)
	      }
	      // erase 'crlistJ'
	      pos = aggs.find( crlistJ ); assert(pos != NULL);
	      aggs.removeAt( pos );
	      delete crlistJ;
	    }
	  }
	  else if(vv > maxv) maxv = vv; 
	}
      }
      nadj[2*leqi] = (int)(1000.0*maxv + .5);
    }
    ierr = Ti->RestoreRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
  }
  
  // aggregate touching aggs w/ strong connection
  for( leqi = 0, geqi = my0Lm ; leqi < lmLoc ; leqi++, geqi++ ){
    if( (crlist=lid_aggList[leqi]) == NULL || nadj[2*leqi] < 0 ) continue;
    // accrete existing agg
    ierr = Ti->GetRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    for( jj = 0 ; jj < nA ; jj++ ){
      const int geqj = Aids[jj], leqj = geqj - my0Lm; vv = fabs(Avals[jj]);
      if( leqj == leqi ) /*void*/;
      else if ( (leqj < 0 || leqj >= lmLoc) && !do_glob ) /*void*/;
      else if( vv >= tol1 ) {
	PromAList<int> *crlistJ; 
	if( leqj >= 0 && leqj < lmLoc ) crlistJ = lid_aggList[leqj]; 
	else crlistJ = ghost_list.find(geqj+1);
	//
	if( crlistJ == NULL ) {	// set 'crlistJ'
	  if( leqj < 0 || leqj >= lmLoc ){
	    int add = ghost_list.add( crlist, geqj+1 ); assert(do_glob&&add==1);
	    if( crlist->count_ == -1 ){
	      crlist->count_ = locAggID++; //flag: has a ghost
	      listID_list.add( crlist, crlist->count_ + 1 );
	    }
	  }
	  else{ lid_aggList[leqj] = crlist; ndone++; } // agg.
	  crlist->addTail(geqj+1); 
	}
	else if( crlist != crlistJ ) { // local
	  //PetscPrintf(PETSC_COMM_SELF,"\t\t\t[%d]%s convert (2) other 'idj' to 'idi'. %d+%d->%d\n",mype_,__FUNCT__,crlist->getCount(),crlistJ->getCount(),crlist->getCount()+crlistJ->getCount());
	  while( !crlistJ->isEmpty() ) {
	    int geqj2 = crlistJ->removeHead() - 1, leqj2 = geqj2 - my0Lm;
	    if( leqj2 < 0 || leqj2 >= lmLoc ){
  	      int add = ghost_list.add( crlist, geqj2+1 ); assert(do_glob&&add==-1);
	      if( crlist->count_ == -1 ){
		crlist->count_ = locAggID++; //flag: has a ghost
		listID_list.add( crlist, crlist->count_ + 1 );
	      }
	    }
	    else lid_aggList[leqj2] = crlist;
	    crlist->addTail( geqj2 + 1 ); // agg. (allready counted)
	  }
	  // erase 'crlistJ'
	  pos = aggs.find( crlistJ ); assert(pos != NULL);
	  aggs.removeAt( pos );
	  delete crlistJ;
	}
      }
    }
    ierr = Ti->RestoreRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
  }

  if(lmLoc>0)PetscPrintf(PETSC_COMM_SELF,"\t[%d] local strong aggs: %d aggs, %d/%d done\n",mype_,aggs.getCount(),ndone,lmLoc);

  // globalize aggs
  if( do_glob ) {
    ierr = GlobalizeAggs( Ti->map_, aggs, lid_aggList, ghost_list, listID_list,
			  &dummy );  CHKERRQ(ierr);
  }

  // grow strong aggs
  for( leqi = 0, geqi = my0Lm ; leqi < lmLoc ; leqi++, geqi++ ){
    if( (crlist=lid_aggList[leqi]) != NULL && nadj[2*leqi] > 0 ) {
      ierr = Ti->GetRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
      // MIS with threshhold
      for( jj = 0 ; jj < nA ; jj++ ){
	const int geqj = Aids[jj], leqj = geqj - my0Lm; 
	if( leqj == leqi ) /*void*/;
	else if ( (leqj < 0 || leqj >= lmLoc) && !do_glob ) /*void*/;
	else if( fabs(Avals[jj]) >= tol2 ) {
	  PromAList<int> *crlistJ; 
	  if( leqj >= 0 && leqj < lmLoc ) crlistJ = lid_aggList[leqj]; 
	  else crlistJ = ghost_list.find(geqj+1); 
	  if( crlistJ == NULL || crlistJ == &dummy ) {
	    // 2nd threshold (local)
	    if((ii=crlist->count_)>=0 && (ii/PROM_MAX_SGAGGS) != mype_ ){
	      // add memo and send to 'proc' -- todo!!!
	      // proc = ii / PROM_MAX_SGAGGS;
	    }
	    else{
	      if( leqj >= 0 && leqj < lmLoc ){
		lid_aggList[leqj] = crlist; ndone++; 
	      }
	      else { // ghost_list.add( crlist, geqj+1 ); -- add here and send message to 'proc' -- todo!!!
		continue;
	      }
	      crlist->addTail(geqj+1); 
	      nadj[2*leqj] = -2; // do not add more with this
	    }
	  }
	}
      }
      ierr = Ti->RestoreRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    }
  }
  //if(lmLoc>0)PetscPrintf(PETSC_COMM_SELF,"\t\t[%d] %d aggs, %d/%d done -- MIS aggs\n",mype_,aggs.getCount(),ndone,lmLoc);
  
  // MIS rest, sort on degree
  qsort( (void*)nadj, lmLoc, 2*sizeof(int), prom_intcompar2 );
  ierr = MIS_CRs( ndone, nadj, aggs, Ti, tol2, do_glob, lid_aggList,
		  ghost_list, listID_list, locAggID, &dummy ); CHKERRQ(ierr);
  assert(ndone==lmLoc);
  
  PetscFree(lid_aggList);  PetscFree(nadj); 
  ghost_list.removeAll(); // could maintain this some more
  
  return 0;
}

static int bi_find( const int *arr, const int sz, const int id ) 
{
  int t, high = sz, low = 0;
  while (high-low > 5) {
    t = (low+high)/2;
    if (arr[t] > id) high = t;
    else             low  = t;
  }
  for (int i=low; i<high; i++) {
    if (arr[i] > id) break;
    if (arr[i] == id) return i;
  }
  return -1; // no find
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::MIS_CRs"
/* Prometheus_LinSysCore::MIS_CRs() ***************************
 *
 *  - 
 *
 *  INPUT:
 *   - map: 
 *   - aggs: (in/out)
 *
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::MIS_CRs(int &ndone, 
				   const int nadj[],
				   PromList<PromAList<int>*> &aggs,
				   const PromMatrix *const Ti,
				   const double tol2, const bool do_glob,
				   PromAList<int> **lid_aggList,
				   PromTable<PromAList<int>*> &ghost_list,
				   PromTable<PromAList<int>*> &listID_list,
				   int &locAggID, PromAList<int> *const dummy)
  const
{
  int ierr,ii,jj,tt,geqi,nA,leqi,*Aids,*lid_hproc,proc,pp,procs[PROM_BSZ];
  const int *cr_geq = Ti->map_.proc_gnode(),np=Ti->map_.comm_.np();
  MPI_Comm wcomm = Ti->map_.comm_.MPIComm(); 
  const int lmLoc=Ti->map_.nlocalEq(),my0Lm=Ti->map_.my0Eq();
  double *Avals,vv; PromAList<int> *crlist;  MPI_Status status;
  const int reqsz = sizeof(MPI_Request)/sizeof(int), tag = Ti->map_.comm_.getNewTag();
  PromList<MPI_Request*> smessages; 
  PromTable<PromList<int>*> proc_list(57); PromList<int> *list;
  // get proc lists
  ierr = PetscMalloc((lmLoc+1)*sizeof(int), &lid_hproc); CHKERRQ(ierr);
  lid_hproc[lmLoc] = -9;
  for( tt = ii = 0 ; ii < lmLoc ; ii++ ){
    geqi = ii + my0Lm; leqi = ii;
    ierr = Ti->GetRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    for( jj = 0, pp = mype_ ; jj < nA ; jj++ ){
      const int geqj = Aids[jj], leqj = geqj-my0Lm; vv = fabs(Avals[jj]);
      if( leqj == leqi ) /* void */;
      else if ( (leqj < 0 || leqj >= lmLoc) && !do_glob ) /*void*/;
      else if( vv >= tol2 ) { // 2nd threshold
	if( leqj >= 0 && leqj < lmLoc ) /*void*/;
	else {
	  PromGrid::FindProc(cr_geq, np, geqj, proc);
	  // store send <gidj, listID>
	  if( (list=proc_list.find(proc+1)) == NULL ) {
	    list = new PromList<int>(16); CHKNEW(list,1);
	    proc_list.add( list, proc+1 );
	    procs[tt++] = proc;
	    if( proc > pp ) pp = proc;
	  }
	}
      }
    }
    ierr = Ti->RestoreRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
    lid_hproc[leqi] = pp;
  }
  const int ptodo = tt;

  // MIS loop
  while( ndone != lmLoc ) {
    //PetscPrintf(PETSC_COMM_WORLD,"[%d] LOOP\n",mype_);
    // MIS with threshhold
    for( ii = 0 ; ii < lmLoc ; ii++ ){
      leqi = nadj[2*ii+1]; 
      if( lid_aggList[leqi] == NULL ) {
	geqi = leqi + my0Lm;
	// see if top
	ierr = Ti->GetRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
	// check for higher proc
	for( jj = tt = 0, pp = mype_ ; jj < nA ; jj++ ){
	  const int geqj = Aids[jj], leqj = geqj - my0Lm; vv = fabs(Avals[jj]);
	  if( leqj == leqi ) /* void */;
	  else if ( (leqj < 0 || leqj >= lmLoc) && !do_glob ) /*void*/;
	  else if( vv >= tol2 ) { // 2nd threshold
	    if( leqj >= 0 && leqj < lmLoc ) { // local see if it can get hit 
	      if( lid_hproc[leqj] > mype_ && lid_aggList[leqj] == NULL ) tt++; 
	    }
	    else {
	      if( ghost_list.find(geqj+1) == NULL ) { // ghost, see if ...
		PromGrid::FindProc(cr_geq,np,geqj,proc); assert(proc!=mype_);
		if( proc > mype_ ){
		  tt++;
		  //PetscPrintf(PETSC_COMM_SELF,"\t\t[%d] %d skips %d, on %d\n",mype_,geqi,geqj,proc);
		  if( proc > pp ) pp = proc;
		}
	      }
	    }
	  }
	}
	if( lid_hproc[leqi] < pp ) lid_hproc[leqi] = pp; // update high
	if( tt == 0 && lid_hproc[leqi] <= mype_ ) {
	  crlist = new PromAList<int>(5); crlist->count_ = -1;
	  aggs.addTail( crlist );
	  lid_aggList[leqi] = crlist; // this is now picked
	  crlist->addTail( geqi + 1 ); ndone++; // agg.
	  // MIS with threshhold
	  for( jj = 0 ; jj < nA ; jj++ ){
	    const int geqj = Aids[jj], leqj = geqj-my0Lm; vv = fabs(Avals[jj]);
	    if( leqj == leqi ) /* void */;
	    else if ( (leqj < 0 || leqj >= lmLoc) && !do_glob ) /*void*/;
	    else if( vv >= tol2 ) { // 2nd threshold
	      if( leqj >= 0 && leqj < lmLoc ) {
		if( lid_aggList[leqj] == NULL ) { 
		  lid_aggList[leqj] = crlist; crlist->addTail(geqj+1); ndone++; // agg.
		}
	      }
	      else {
		if( ghost_list.find(geqj+1) == NULL ) {
		  if( crlist->count_ == -1 ){ 
		    crlist->count_ = locAggID++;
		    listID_list.add( crlist, crlist->count_ + 1 );
		  }
		  // add it in
		  ghost_list.add( crlist, geqj+1 ); crlist->addTail(geqj+1); 
		  // store send <gidj, geqi, listID>
		  PromGrid::FindProc(cr_geq, np, geqj, proc); assert(proc<mype_);
		  list = proc_list.find( proc + 1 ); assert(list != NULL);
		  list->addTail( geqj + 1 ); list->addTail( geqi + 1 ); 
		  list->addTail( crlist->count_ + 1 );
		  //PetscPrintf(PETSC_COMM_SELF,"\t\t[%d] %d adds ghost %d, on %d\n",mype_,geqi,geqj,proc);
		}
	      }
	    }
	  }
	  //PetscPrintf(PETSC_COMM_SELF,"\t\t\t[%d] MIS agg(%d), %d eqs\n",mype_,nadj[2*ii],crlist->getCount());
	}
	ierr = Ti->RestoreRow( geqi, &nA, &Aids, &Avals );   CHKERRQ(ierr);
      }
    }
    // send
    TablePos tpos = proc_list.getHeadPosition();
    while( tpos ) {
      list = proc_list.getNext( tpos, &proc ); proc--;
      int tt = list->getCount(); 
      MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq);
      CHKERRQ(ierr);
      smessages.addTail( sreq );
      int *buff = (int*)sreq + reqsz, *pb = buff;
      while( !list->isEmpty() ){ 
	int gidj = list->removeHead()-1, gidi = list->removeHead()-1; 
	ii = list->removeHead() - 1;
	*pb++ = gidj; *pb++ = gidi; *pb++ = ii;
      }
      if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
      // send(tag) *< gidj, geqi, listID >
      MPI_Isend( buff, tt, MPI_INT, proc, tag, wcomm, sreq );
    }
    // recv
    for( ii = 0 ; ii < ptodo ; ii++ ) {
      MPI_Probe( procs[ii], tag, wcomm, &status );
      MPI_Get_count( &status, MPI_INT, &tt ); 
      if( tt == 0 ) {
	MPI_Recv( NULL, 0, MPI_INT, status.MPI_SOURCE, tag, wcomm, &status);
      }
      else {
	int *buff,*pb,kk;
	ierr = PetscMalloc(tt*sizeof(int),&buff); CHKERRQ(ierr); pb = buff;
	// recv(tag) *< gidi, geqj, listID >
	MPI_Recv(buff, tt, MPI_INT, status.MPI_SOURCE, tag, wcomm, &status);
	kk = tt/3; assert(tt%3 == 0);
	while( kk-- ) {
	  const int geqi = *pb++, leqi = geqi - my0Lm; 
	  const int geqj = *pb++, ltid = *pb++; assert(leqi>=0 && leqi<lmLoc);
	  assert(lid_aggList[leqi] == NULL);
	  crlist = listID_list.find( ltid + 1 );
	  if( crlist == NULL ) {
	    crlist = new PromAList<int>(5); crlist->count_ = ltid;
	    aggs.addTail( crlist );
	    listID_list.add( crlist, ltid + 1 );
	  }
	  lid_aggList[leqi] = crlist; crlist->addTail(geqi+1); ndone++; // agg.
	  ghost_list.add( crlist, geqj+1 ); crlist->addTail(geqj+1); 
	}
	if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
	PetscFree(buff); 
      }
    }
    // clean isend messages
    while( !smessages.isEmpty() ) {
      MPI_Request *sreq = smessages.removeHead();
      MPI_Wait( sreq, &status );    PetscFree( sreq );
    }
  }

  // delte list
  while( !proc_list.isEmpty() ) { 
    list = proc_list.removeHead(); delete list;
  }
  assert(lid_hproc[lmLoc] == -9); PetscFree(lid_hproc); 
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::GlobalizeAggs"
/* Prometheus_LinSysCore::GlobalizeAggs() ***************************
 *
 *  - 
 *
 *  INPUT:
 *   - map: 
 *   - aggs: (in/out)
 *
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::GlobalizeAggs(const PromMap &map, 
					 PromList<PromAList<int>*> &aggs,
					 PromAList<int> **lid_aggList,
					 PromTable<PromAList<int>*> &ghost_list,
					 PromTable<PromAList<int>*> &listID_list,
					 PromAList<int> *const dummy) const
{
  int ierr,ii,jj,tt,lcid,proc,*bf,*buff1,*buff2,*listIDs,*temp,leq,geq,id;
  const int lmLoc=map.nlocalEq(),my0Lm=map.my0Eq(),np=map.comm_.np();
  MPI_Comm wcomm = map.comm_(); PromAList<int> *crlist;
  PromList<PromAList<int>*> bound_aggs; Position pos; 
  int buf[PROM_BSZ]; PromAList<int> *aggArr[PROM_BSZ];

  pos = aggs.getHeadPosition(); ii = 0;
  while( pos ){
    crlist = aggs.getNext( pos );
    if( crlist->count_ == -1 ) continue;  
    Position pos2 = crlist->getHeadPosition(); tt = 0; // find ghost
    while( pos2 ) {
      leq = crlist->getNext( pos2 ) - my0Lm - 1;
      if( leq >= 0 && leq < lmLoc ) { // local
	buf[2*ii] = leq + my0Lm; buf[2*ii+1] = crlist->count_;
	ii++; assert(ii<PROM_BSZ/2); // count local in boundary aggregate
      }
      else tt++;
    }
    assert( tt > 0 );  // boundary agg
    bound_aggs.addTail( crlist );
  }
  
  // bufers
  const int nlocBnd = ii, nBAggs = bound_aggs.getCount();
  MPI_Allreduce( &ii, &tt, 1, MPI_INT, MPI_SUM, wcomm );// global!
  const int nglobBnd = tt; ii = 3*(nglobBnd+1); if(ii<np)ii=np;
  if(nglobBnd == 0) return 0; // nothing to do
  ierr = PetscMalloc(ii*sizeof(int), &bf); CHKERRQ(ierr);
  buff1 = bf + 2*(nglobBnd+1); buff2 = bf + nglobBnd+1;

  // form 'aggArr'
  assert(nBAggs<PROM_BSZ);
  for(ii=0;ii<nBAggs;ii++)aggArr[ii] = NULL;
  pos = bound_aggs.getHeadPosition(); 
  while( pos ){
    crlist = bound_aggs.getNext( pos );
    ii = crlist->count_%PROM_MAX_SGAGGS; assert(ii<nBAggs);// local ID
    assert(aggArr[ii] == NULL); aggArr[ii] = crlist;
  }

  {   // collect gids
    qsort( (void*)buf, nlocBnd, 2*sizeof(int), prom_intcompar );
    int sendcount = 2*nlocBnd, recvcounts[PROM_BSZ], displs[PROM_BSZ];
    MPI_Allgather( &sendcount, 1, MPI_INT, recvcounts, 1, MPI_INT, wcomm );
    for(ii=0,displs[0]=0;ii<np-1;ii++)displs[ii+1]=displs[ii]+recvcounts[ii];
    MPI_Allgatherv(buf,sendcount,MPI_INT,bf,recvcounts,displs,MPI_INT,wcomm);
  }
  for(ii=jj=0;ii<nglobBnd;ii++,jj+=2){bf[ii]=bf[jj];buff1[ii]=bf[jj+1];}
  bf[nglobBnd] = buff1[nglobBnd] = buff2[nglobBnd] = -999;
  const int *const gids = bf;

  listIDs = buff1; //current data
  if( nlocBnd ) id = bi_find( gids, nglobBnd, buf[0] ); 
  else id = -1; 
  const int my0BID = id;  assert(nBAggs<PROM_BSZ);

  // collect highest 
  while(1){
    pos = bound_aggs.getHeadPosition(); int ndone = 0;
    while( pos ){
      crlist = bound_aggs.getNext(pos); assert(crlist->count_ != -1);
      // get highest ID, and collect gids
      Position poses[PROM_BSZ]; 
      Position pos2 = crlist->getHeadPosition(); 
      for( ii = 0, tt = -1; pos2 ; ii++ ) {
	poses[ii] = pos2; // save for removeal
	geq = crlist->getNext( pos2 ) - 1; leq = geq - my0Lm;
	if( leq < 0 || leq >= lmLoc ) { // ghost -- bisection search
	  id = bi_find( gids, nglobBnd, geq ); assert(id>=0 && gids[id]==geq);
	}
	else {                          // ghost -- bisection search
	  assert(my0BID>=0);
	  id = bi_find( &gids[my0BID], nlocBnd, geq ); assert(id>=0);
	  id += my0BID; assert(gids[id] == geq);
	}
	buf[ii] = id; // collect these geqs
	if( listIDs[id] > tt ) tt = listIDs[id]; // get max
      }
      assert(ii<PROM_BSZ);
      const int highID = tt, nn = ii;

      // set highest listID
      for( ii = 0 ; ii < nn ; ii++ ){
	id = buf[ii]; 
	if( listIDs[id] < highID ) { 
	  listIDs[id] = highID; ndone++; // changed one
	  // remove from old
	  crlist->removeAt( poses[ii] ) - 1;
	  geq = gids[id]; leq = geq - my0Lm;
	  proc = highID/PROM_MAX_SGAGGS; PromAList<int> *newlist;
	  // get 'newlist'
	  if( proc == mype_ ) { 
	    lcid = highID%PROM_MAX_SGAGGS;                assert(lcid<nBAggs);
	    newlist = aggArr[lcid];
	  }
	  else { // local moved to other proc
	    newlist = listID_list.find( highID + 1 );
	    if( newlist == NULL && leq >= 0 && leq < lmLoc ) {
	      newlist = new PromAList<int>(5); newlist->count_ = highID;
	      aggs.addTail( newlist );
	      listID_list.add( newlist, highID + 1 );
	    }
	  }
	  // set new list
	  if( leq >= 0 && leq < lmLoc ) lid_aggList[leq] = newlist;
	  else ghost_list.add( (newlist==NULL) ? dummy : newlist, geq+1 );
	  //
	  if(newlist != NULL) newlist->addTail( geq + 1 );
	}
	else assert(listIDs[id] == highID);
      }
    }
    // global!
    temp = (listIDs == buff1) ? buff2 : buff1;
    MPI_Allreduce( listIDs, temp, nglobBnd, MPI_INT, MPI_MAX, wcomm );
    listIDs = temp;

    // set locals that were moved
    for( jj = 0, ii = my0BID ; jj < nlocBnd ; ii++, jj++ ) {
      tt = listIDs[ii]; proc = tt/PROM_MAX_SGAGGS; 
      geq = gids[ii]; leq = geq - my0Lm; assert(leq>=0 && leq<lmLoc);
      if( proc == mype_ ) { // one that I need
	lcid = tt%PROM_MAX_SGAGGS;  assert(lcid<nBAggs);
	crlist = aggArr[lcid]; assert(crlist != NULL && crlist->count_==tt);
	if( lid_aggList[leq] != crlist ) {
	  PromAList<int> *crlistQ = lid_aggList[leq];  assert(crlistQ!=NULL);
	  while( !crlistQ->isEmpty() ) {
	    int geqj2 = crlistQ->removeHead() - 1, leqj2 = geqj2 - my0Lm;
	    if( leqj2 < 0 || leqj2 >= lmLoc ){
	      ghost_list.add( crlist, geqj2+1 ); 
	      assert(crlist->count_ != -1); ndone++; // new agg.
	    }
	    else lid_aggList[leqj2] = crlist;
	    crlist->addTail( geqj2 + 1 ); 
	  }
	  // erase 'crlistQ'
	  pos = aggs.find( crlistQ ); assert(pos != NULL);
	  aggs.removeAt( pos );
	  lcid = crlistQ->count_%PROM_MAX_SGAGGS;  assert(lcid<nBAggs);
	  aggArr[lcid] = NULL;
	  delete crlistQ;
	}
      }
      else if( lid_aggList[leq]->count_ != tt ) {
	crlist = lid_aggList[leq]; assert(crlist!=NULL); // moved off
	pos = crlist->find( geq + 1 ); assert(pos!=NULL);
	crlist->removeAt( pos );
	PromAList<int> *crlistQ = listID_list.find( tt + 1 );
	// add to other
	if( crlistQ == NULL ) {
	  crlistQ = new PromAList<int>(5); crlistQ->count_ = tt;
	  aggs.addTail( crlistQ ); // new other
	  listID_list.add( crlistQ, tt + 1 );
	}
	else assert(crlistQ != crlist);
	lid_aggList[leq] = crlistQ;
	crlistQ->addTail( geq + 1 ); // agg. (allready counted)
      }
      else assert(lid_aggList[leq]->count_/PROM_MAX_SGAGGS != mype_);
    }
    // global!
    MPI_Allreduce( &ndone, &ii, 1, MPI_INT, MPI_SUM, wcomm );
    //PetscPrintf(PETSC_COMM_WORLD,"\t\t\t\t\t[%d] LOOP %d done\n",mype_,ii);
    if(ii == 0) break;
  }
  temp = (listIDs == buff1) ? buff2 : buff1;

  // add to lists (not supper scalable)
  for( ii = 0 ; ii < nglobBnd ; ii++ ) {
    tt = listIDs[ii]; proc = tt/PROM_MAX_SGAGGS; 
    geq = gids[ii]; leq = geq - my0Lm;
    if( proc == mype_ ) { // one that I need
      lcid = tt%PROM_MAX_SGAGGS;  assert(lcid<nBAggs);
      crlist = aggArr[lcid]; assert(crlist != NULL); assert(crlist->count_==tt);
      if( leq < 0 || leq >= lmLoc ) { // a ghost add it in
	PromAList<int> *crlistJ = ghost_list.find(geq+1);  
	if( crlistJ != crlist ) {
	  if( crlistJ != dummy && crlistJ != NULL){
	    pos = crlistJ->find(geq+1); crlistJ->removeAt(pos);
	  }
	  crlist->addTail( geq+ 1 ); 
	}
      }
      else if( lid_aggList[leq] != crlist ){ // local that is moved local?
	assert( lid_aggList[leq] != NULL ); assert(lid_aggList[leq]==crlist);
      }
    }
    else if( leq >= 0 && leq < lmLoc ) { // local that got moved
      crlist = lid_aggList[leq]; assert(crlist!=NULL);
      if( crlist->count_ != tt && crlist->count_/PROM_MAX_SGAGGS == mype_ ) {
	lcid = crlist->count_%PROM_MAX_SGAGGS;  assert(lcid<nBAggs);
	PromAList<int> *crlistQ = aggArr[lcid]; 
	pos = crlistQ->find(geq+1); assert(pos!=NULL);
	crlistQ->removeAt(pos);
      }
      else assert(crlist->count_ == tt);
    }
    else { // ghost moved off
      PromAList<int> *crlistJ = ghost_list.find( geq + 1 ); 
      if( crlistJ != NULL && crlistJ != dummy ) {
	PromAList<int> *newlist = listID_list.find( tt + 1 );
	if( crlistJ != newlist ) { // lost a ghost - delete locally
	  pos = crlistJ->find(geq+1); assert(pos!=NULL);
	  crlistJ->removeAt( pos );
	  if( newlist != NULL ) ghost_list.add( newlist, geq+1 ); // ghost
	  else ghost_list.add( dummy, geq+1 ); 
	}
      }
    }
  }
  assert(bf[nglobBnd] == -999);assert(buff1[nglobBnd] == -999);
  assert(buff2[nglobBnd] == -999);
  PetscFree(bf); 

  return 0;
}
